@@ -1,5 +1,5 @@
 /* Three way file comparison program (diff3) for Project GNU.
-   Copyright (C) 1988, 1989, 1992, 1993, 1994, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1988, 1989, 1992, 1993, 1994, 1997, 1998 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,9 +11,7 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   */
 
 /* Written by Randy Smith */
 /* Librarification by Tim Pierce */
@@ -22,6 +20,7 @@
 #include <stdio.h>
 #include <setjmp.h>
 #include "getopt.h"
+#include "diffrun.h"
 
 /* diff3.c has a real initialize_main function. */
 #ifdef initialize_main
@@ -30,6 +29,18 @@
 
 extern char const diff_version_string[];
 
+extern FILE *outfile;
+
+extern const struct diff_callbacks *callbacks;
+
+void write_output PARAMS((char const *, size_t));
+void printf_output PARAMS((char const *, ...))
+#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 6)
+     __attribute__ ((__format__ (__printf__, 1, 2)))
+#endif
+     ;
+void flush_output PARAMS((void));
+
 /*
  * Internal data structures and macros for the diff3 program; includes
  * data structures for both diff3 diffs and normal diffs.
@@ -186,26 +197,27 @@ static char *scan_diff_line PARAMS((char *, char **, size_t *, char *, int));
 static enum diff_type process_diff_control PARAMS((char **, struct diff_block *));
 static int compare_line_list PARAMS((char * const[], size_t const[], char * const[], size_t const[], int));
 static int copy_stringlist PARAMS((char * const[], size_t const[], char *[], size_t[], int));
-static int dotlines PARAMS((FILE *, struct diff3_block *, int));
-static int output_diff3_edscript PARAMS((FILE *, struct diff3_block *, int const[3], int const[3], char const *, char const *, char const *));
-static int output_diff3_merge PARAMS((FILE *, FILE *, struct diff3_block *, int const[3], int const[3], char const *, char const *, char const *));
+static int dotlines PARAMS((struct diff3_block *, int));
+static int output_diff3_edscript PARAMS((struct diff3_block *, int const[3], int const[3], char const *, char const *, char const *));
+static int output_diff3_merge PARAMS((FILE *, struct diff3_block *, int const[3], int const[3], char const *, char const *, char const *));
 static size_t myread PARAMS((int, char *, size_t));
 static struct diff3_block *create_diff3_block PARAMS((int, int, int, int, int, int));
 static struct diff3_block *make_3way_diff PARAMS((struct diff_block *, struct diff_block *));
 static struct diff3_block *reverse_diff3_blocklist PARAMS((struct diff3_block *));
 static struct diff3_block *using_to_diff3_block PARAMS((struct diff_block *[2], struct diff_block *[2], int, int, struct diff3_block const *));
-static struct diff_block *process_diff PARAMS((char const *, char const *, struct diff_block **));
+static struct diff_block *process_diff PARAMS((char const *, char const *, struct diff_block **, char **));
 static void check_output PARAMS((FILE *));
 static void diff3_fatal PARAMS((char const *));
-static void output_diff3 PARAMS((FILE *, struct diff3_block *, int const[3], int const[3]));
+static void output_diff3 PARAMS((struct diff3_block *, int const[3], int const[3]));
 static void diff3_perror_with_exit PARAMS((char const *));
 static int try_help PARAMS((char const *));
-static void undotlines PARAMS((FILE *, int, int, int));
+static void undotlines PARAMS((int, int, int));
 static void usage PARAMS((void));
 static void initialize_main PARAMS((int *, char ***));
+static void free_diff_blocks PARAMS((struct diff_block *));
+static void free_diff3_blocks PARAMS((struct diff3_block *));
 
 /* Functions provided in libdiff.a or other external sources. */
-int diff_run PARAMS((int, char **, char *));
 VOID *xmalloc PARAMS((size_t));
 VOID *xrealloc PARAMS((VOID *, size_t));
 void perror_with_name PARAMS((char const *));
@@ -236,10 +248,11 @@ static struct option const longopts[] =
  * combines the two diffs, and outputs them.
  */
 int
-diff3_run (argc, argv, outfile)
+diff3_run (argc, argv, out, callbacks_arg)
      int argc;
      char **argv;
-     char *outfile;
+     char *out;
+     const struct diff_callbacks *callbacks_arg;
 {
   int c, i;
   int mapping[3];
@@ -248,6 +261,7 @@ diff3_run (argc, argv, outfile)
   int conflicts_found;
   int status;
   struct diff_block *thread0, *thread1, *last_block;
+  char *content0, *content1;
   struct diff3_block *diff3;
   int tag_count = 0;
   char *tag_strings[3];
@@ -255,7 +269,9 @@ diff3_run (argc, argv, outfile)
   char **file;
   struct stat statb;
   int optind_old;
-  FILE *outstream;
+  int opened_file = 0;
+
+  callbacks = callbacks_arg;
 
   initialize_main (&argc, &argv);
 
@@ -300,11 +316,19 @@ diff3_run (argc, argv, outfile)
 	  tab_align_flag = 1;
 	  break;
 	case 'v':
-	  printf ("diff3 - GNU diffutils version %s\n", diff_version_string);
+	  if (callbacks && callbacks->write_stdout)
+	    {
+	      (*callbacks->write_stdout) ("diff3 - GNU diffutils version ");
+	      (*callbacks->write_stdout) (diff_version_string);
+	      (*callbacks->write_stdout) ("\n");
+	    }
+	  else
+	    printf ("diff3 - GNU diffutils version %s\n", diff_version_string);
 	  return 0;
 	case 129:
 	  usage ();
-	  check_output (stdout);
+	  if (! callbacks || ! callbacks->write_stdout)
+	    check_output (stdout);
 	  return 0;
 	case 'L':
 	  /* Handle up to three -L options.  */
@@ -380,22 +404,33 @@ diff3_run (argc, argv, outfile)
           }
 	else if (S_ISDIR(statb.st_mode))
 	  {
-	    fprintf (stderr, "%s: %s: Is a directory\n",
-		     diff_program_name, file[i]);
+	    diff_error ("%s: Is a directory", file[i], 0);
 	    return 2;
 	  }
       }
 
-  if (outfile == NULL)
-    outstream = stdout;
+  if (callbacks && callbacks->write_output)
+    {
+      if (out != NULL)
+	{
+	  diff_error ("write callback with output file", 0, 0);
+	  return 2;
+	}
+    }
   else
     {
-      outstream = fopen (outfile, "w");
-      if (outstream == NULL)
-        {
-	  perror_with_name ("could not open output file");
-	  return 2;
-        }
+      if (out == NULL)
+	outfile = stdout;
+      else
+	{
+	  outfile = fopen (out, "w");
+	  if (outfile == NULL)
+	    {
+	      perror_with_name ("could not open output file");
+	      return 2;
+	    }
+	  opened_file = 1;
+	}
     }
 
   /* Set the jump buffer, so that diff may abort execution without
@@ -405,39 +440,52 @@ diff3_run (argc, argv, outfile)
       return status;
 
   commonname = file[rev_mapping[FILEC]];
-  thread1 = process_diff (file[rev_mapping[FILE1]], commonname, &last_block);
+  thread1 = process_diff (file[rev_mapping[FILE1]], commonname, &last_block,
+			  &content1);
   if (thread1)
     for (i = 0; i < 2; i++)
       {
 	horizon_lines = max (horizon_lines, D_NUMLINES (thread1, i));
 	horizon_lines = max (horizon_lines, D_NUMLINES (last_block, i));
       }
-  thread0 = process_diff (file[rev_mapping[FILE0]], commonname, &last_block);
+  thread0 = process_diff (file[rev_mapping[FILE0]], commonname, &last_block,
+			  &content0);
   diff3 = make_3way_diff (thread0, thread1);
   if (edscript)
     conflicts_found
-      = output_diff3_edscript (outstream, diff3, mapping, rev_mapping,
+      = output_diff3_edscript (diff3, mapping, rev_mapping,
 			       tag_strings[0], tag_strings[1], tag_strings[2]);
   else if (merge)
     {
-      if (! freopen (file[rev_mapping[FILE0]], "r", stdin))
+      FILE *mfp = fopen (file[rev_mapping[FILE0]], "r");
+      if (! mfp)
 	diff3_perror_with_exit (file[rev_mapping[FILE0]]);
-      conflicts_found
-	= output_diff3_merge (stdin, outstream, diff3, mapping, rev_mapping,
+      conflicts_found = output_diff3_merge (mfp, diff3, mapping, rev_mapping,
 			      tag_strings[0], tag_strings[1], tag_strings[2]);
-      if (ferror (stdin))
+      if (ferror (mfp))
 	diff3_fatal ("read error");
+      if (fclose(mfp) != 0)
+	perror_with_name (file[rev_mapping[FILE0]]);
     }
   else
     {
-      output_diff3 (outstream, diff3, mapping, rev_mapping);
+      output_diff3 (diff3, mapping, rev_mapping);
       conflicts_found = 0;
     }
 
-  check_output (outstream);
-  if (outstream != stdout)
-    if (fclose (outstream) != 0)
-	perror ("close error on output file");
+  free(content0);
+  free(content1);
+  free_diff_blocks(thread0);
+  free_diff_blocks(thread1);
+  free_diff3_blocks(diff3);
+
+  if (! callbacks || ! callbacks->write_output)
+    check_output (outfile);
+
+  if (opened_file)
+    if (fclose (outfile) != 0)
+	perror_with_name ("close error on output file");
+
   return conflicts_found;
 }
 
@@ -446,9 +494,8 @@ try_help (reason)
      char const *reason;
 {
   if (reason)
-    fprintf (stderr, "%s: %s\n", diff_program_name, reason);
-  fprintf (stderr, "%s: Try `%s --help' for more information.\n",
-	   diff_program_name, diff_program_name);
+    diff_error ("%s", reason, 0);
+  diff_error ("Try `%s --help' for more information.", diff_program_name, 0);
   return 2;
 }
 
@@ -466,25 +513,52 @@ check_output (stream)
 static void
 usage ()
 {
-  printf ("Usage: %s [OPTION]... MYFILE OLDFILE YOURFILE\n\n", diff_program_name);
+  if (callbacks && callbacks->write_stdout)
+    {
+      (*callbacks->write_stdout) ("Usage: ");
+      (*callbacks->write_stdout) (diff_program_name);
+      (*callbacks->write_stdout) (" [OPTION]... MYFILE OLDFILE YOURFILE\n\n");
 
-  printf ("%s", "\
+      (*callbacks->write_stdout) ("\
   -e  --ed  Output unmerged changes from OLDFILE to YOURFILE into MYFILE.\n\
   -E  --show-overlap  Output unmerged changes, bracketing conflicts.\n\
   -A  --show-all  Output all changes, bracketing conflicts.\n\
   -x  --overlap-only  Output overlapping changes.\n\
   -X  Output overlapping changes, bracketing them.\n\
   -3  --easy-only  Output unmerged nonoverlapping changes.\n\n");
-  printf ("%s", "\
+      (*callbacks->write_stdout) ("\
   -m  --merge  Output merged file instead of ed script (default -A).\n\
   -L LABEL  --label=LABEL  Use LABEL instead of file name.\n\
   -i  Append `w' and `q' commands to ed scripts.\n\
   -a  --text  Treat all files as text.\n\
   -T  --initial-tab  Make tabs line up by prepending a tab.\n\n");
-  printf ("%s", "\
+      (*callbacks->write_stdout) ("\
   -v  --version  Output version info.\n\
   --help  Output this help.\n\n");
-  printf ("If a FILE is `-', read standard input.\n");
+      (*callbacks->write_stdout) ("If a FILE is `-', read standard input.\n");
+    }
+  else
+    {
+      printf ("Usage: %s [OPTION]... MYFILE OLDFILE YOURFILE\n\n", diff_program_name);
+
+      printf ("%s", "\
+  -e  --ed  Output unmerged changes from OLDFILE to YOURFILE into MYFILE.\n\
+  -E  --show-overlap  Output unmerged changes, bracketing conflicts.\n\
+  -A  --show-all  Output all changes, bracketing conflicts.\n\
+  -x  --overlap-only  Output overlapping changes.\n\
+  -X  Output overlapping changes, bracketing them.\n\
+  -3  --easy-only  Output unmerged nonoverlapping changes.\n\n");
+      printf ("%s", "\
+  -m  --merge  Output merged file instead of ed script (default -A).\n\
+  -L LABEL  --label=LABEL  Use LABEL instead of file name.\n\
+  -i  Append `w' and `q' commands to ed scripts.\n\
+  -a  --text  Treat all files as text.\n\
+  -T  --initial-tab  Make tabs line up by prepending a tab.\n\n");
+      printf ("%s", "\
+  -v  --version  Output version info.\n\
+  --help  Output this help.\n\n");
+      printf ("If a FILE is `-', read standard input.\n");
+    }
 }
 
 /*
@@ -975,19 +1049,19 @@ compare_line_list (list1, lengths1, list2, lengths2, nl)
 extern char **environ;
 
 static struct diff_block *
-process_diff (filea, fileb, last_block)
+process_diff (filea, fileb, last_block, diff_contents)
      char const *filea, *fileb;
      struct diff_block **last_block;
+     char **diff_contents;
 {
-  char *diff_contents;
   char *diff_limit;
   char *scan_diff;
   enum diff_type dt;
   int i;
   struct diff_block *block_list, **block_list_end, *bptr;
 
-  diff_limit = read_diff (filea, fileb, &diff_contents);
-  scan_diff = diff_contents;
+  diff_limit = read_diff (filea, fileb, diff_contents);
+  scan_diff = *diff_contents;
   block_list_end = &block_list;
   bptr = 0; /* Pacify `gcc -W'.  */
 
@@ -1000,12 +1074,13 @@ process_diff (filea, fileb, last_block)
       dt = process_diff_control (&scan_diff, bptr);
       if (dt == ERROR || *scan_diff != '\n')
 	{
-	  fprintf (stderr, "%s: diff error: ", diff_program_name);
-	  do
-	    {
-	      putc (*scan_diff, stderr);
-	    }
-	  while (*scan_diff++ != '\n');
+	  char *serr;
+
+	  for (serr = scan_diff; *serr != '\n'; serr++)
+	    ;
+	  *serr = '\0';
+	  diff_error ("diff error: %s", scan_diff, 0);
+	  *serr = '\n';
 	  DIFF3_ABORT (2);
 	}
       scan_diff++;
@@ -1174,6 +1249,10 @@ read_diff (filea, fileb, output_placement)
   size_t bytes, current_chunk_size, total;
   int fd, wstatus;
   struct stat pipestat;
+  FILE *outfile_hold;
+  const struct diff_callbacks *callbacks_hold;
+  struct diff_callbacks my_callbacks;
+  struct diff_callbacks *my_callbacks_arg;
 
   /* 302 / 1000 is log10(2.0) rounded up.  Subtract 1 for the sign bit;
      add 1 for integer division truncation; add 1 more for a minus sign.  */
@@ -1196,7 +1275,30 @@ read_diff (filea, fileb, output_placement)
   *ap = 0;
 
   diffout = tmpnam(NULL);
-  wstatus = diff_run (ap - argv, (char **) argv, diffout);
+
+  outfile_hold = outfile;
+  callbacks_hold = callbacks;
+
+  /* We want to call diff_run preserving any stdout and stderr
+     callbacks, but discarding any callbacks to handle file output,
+     since we want the file output to go to our temporary file.
+     FIXME: We should use callbacks to just read it into a memory
+     buffer; that's we do with the temporary file just below anyhow.  */
+  if (callbacks == NULL)
+    my_callbacks_arg = NULL;
+  else
+    {
+      my_callbacks = *callbacks;
+      my_callbacks.write_output = NULL;
+      my_callbacks.flush_output = NULL;
+      my_callbacks_arg = &my_callbacks;
+    }
+
+  wstatus = diff_run (ap - argv, (char **) argv, diffout, my_callbacks_arg);
+
+  outfile = outfile_hold;
+  callbacks = callbacks_hold;
+
   if (wstatus == 2)
     diff3_fatal ("subsidiary diff failed");
 
@@ -1271,17 +1373,25 @@ scan_diff_line (scan_ptr, set_start, set_length, limit, leadingchar)
   *set_length = line_ptr - *set_start;
   if (line_ptr < limit && *line_ptr == '\\')
     {
-      if (edscript)
-	fprintf (stderr, "%s:", diff_program_name);
-      else
-	--*set_length;
-      line_ptr++;
-      do
+      if (! edscript)
 	{
-	  if (edscript)
-	    putc (*line_ptr, stderr);
+	  --*set_length;
+	  line_ptr++;
+	  while (*line_ptr++ != '\n')
+	    ;
+	}
+      else
+	{
+	  char *serr;
+
+	  line_ptr++;
+	  serr = line_ptr;
+	  while (*line_ptr++ != '\n')
+	    ;
+	  line_ptr[-1] = '\0';
+	  diff_error ("%s", serr, 0);
+	  line_ptr[-1] = '\n';
 	}
-      while (*line_ptr++ != '\n');
     }
 
   return line_ptr;
@@ -1299,8 +1409,7 @@ scan_diff_line (scan_ptr, set_start, set_length, limit, leadingchar)
  * REV_MAPPING is the inverse of MAPPING.
  */
 static void
-output_diff3 (outputfile, diff, mapping, rev_mapping)
-     FILE *outputfile;
+output_diff3 (diff, mapping, rev_mapping)
      struct diff3_block *diff;
      int const mapping[3], rev_mapping[3];
 {
@@ -1337,7 +1446,7 @@ output_diff3 (outputfile, diff, mapping, rev_mapping)
 	default:
 	  diff3_fatal ("internal error: invalid diff type passed to output");
 	}
-      fprintf (outputfile, "====%s\n", x);
+      printf_output ("====%s\n", x);
 
       /* Go 0, 2, 1 if the first and third outputs are equivalent.  */
       for (i = 0; i < 3;
@@ -1348,17 +1457,17 @@ output_diff3 (outputfile, diff, mapping, rev_mapping)
 	    lowt = D_LOWLINE (ptr, realfile),
 	    hight = D_HIGHLINE (ptr, realfile);
 
-	  fprintf (outputfile, "%d:", i + 1);
+	  printf_output ("%d:", i + 1);
 	  switch (lowt - hight)
 	    {
 	    case 1:
-	      fprintf (outputfile, "%da\n", lowt - 1);
+	      printf_output ("%da\n", lowt - 1);
 	      break;
 	    case 0:
-	      fprintf (outputfile, "%dc\n", lowt);
+	      printf_output ("%dc\n", lowt);
 	      break;
 	    default:
-	      fprintf (outputfile, "%d,%dc\n", lowt, hight);
+	      printf_output ("%d,%dc\n", lowt, hight);
 	      break;
 	    }
 
@@ -1369,14 +1478,14 @@ output_diff3 (outputfile, diff, mapping, rev_mapping)
 	      line = 0;
 	      do
 		{
-		  fprintf (outputfile, line_prefix);
+		  printf_output (line_prefix);
 		  cp = D_RELNUM (ptr, realfile, line);
 		  length = D_RELLEN (ptr, realfile, line);
-		  fwrite (cp, sizeof (char), length, outputfile);
+		  write_output (cp, length);
 		}
 	      while (++line < hight - lowt + 1);
 	      if (cp[length - 1] != '\n')
-		fprintf (outputfile, "\n\\ No newline at end of file\n");
+		printf_output ("\n\\ No newline at end of file\n");
 	    }
 	}
     }
@@ -1384,12 +1493,11 @@ output_diff3 (outputfile, diff, mapping, rev_mapping)
 
 
 /*
- * Output to OUTPUTFILE the lines of B taken from FILENUM.
+ * Output the lines of B taken from FILENUM.
  * Double any initial '.'s; yield nonzero if any initial '.'s were doubled.
  */
 static int
-dotlines (outputfile, b, filenum)
-     FILE *outputfile;
+dotlines (b, filenum)
      struct diff3_block *b;
      int filenum;
 {
@@ -1404,10 +1512,9 @@ dotlines (outputfile, b, filenum)
       if (line[0] == '.')
 	{
 	  leading_dot = 1;
-	  fprintf (outputfile, ".");
+	  write_output (".", 1);
 	}
-      fwrite (line, sizeof (char),
-	      D_RELLEN (b, filenum, i), outputfile);
+      write_output (line, D_RELLEN (b, filenum, i));
     }
 
   return leading_dot;
@@ -1419,16 +1526,15 @@ dotlines (outputfile, b, filenum)
  * starting with line START and continuing for NUM lines.
  */
 static void
-undotlines (outputfile, leading_dot, start, num)
-     FILE *outputfile;
+undotlines (leading_dot, start, num)
      int leading_dot, start, num;
 {
-  fprintf (outputfile, ".\n");
+  write_output (".\n", 2);
   if (leading_dot)
     if (num == 1)
-      fprintf (outputfile, "%ds/^\\.//\n", start);
+      printf_output ("%ds/^\\.//\n", start);
     else
-      fprintf (outputfile, "%d,%ds/^\\.//\n", start, start + num - 1);
+      printf_output ("%d,%ds/^\\.//\n", start, start + num - 1);
 }
 
 /*
@@ -1454,9 +1560,7 @@ undotlines (outputfile, leading_dot, start, num)
  */
 
 static int
-output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
-		       file0, file1, file2)
-     FILE *outputfile;
+output_diff3_edscript (diff, mapping, rev_mapping, file0, file1, file2)
      struct diff3_block *diff;
      int const mapping[3], rev_mapping[3];
      char const *file0, *file1, *file2;
@@ -1491,22 +1595,22 @@ output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
 
 	  /* Mark end of conflict.  */
 
-	  fprintf (outputfile, "%da\n", D_HIGHLINE (b, mapping[FILE0]));
+	  printf_output ("%da\n", D_HIGHLINE (b, mapping[FILE0]));
 	  leading_dot = 0;
 	  if (type == DIFF_ALL)
 	    {
 	      if (show_2nd)
 		{
 		  /* Append lines from FILE1.  */
-		  fprintf (outputfile, "||||||| %s\n", file1);
-		  leading_dot = dotlines (outputfile, b, mapping[FILE1]);
+		  printf_output ("||||||| %s\n", file1);
+		  leading_dot = dotlines (b, mapping[FILE1]);
 		}
 	      /* Append lines from FILE2.  */
-	      fprintf (outputfile, "=======\n");
-	      leading_dot |= dotlines (outputfile, b, mapping[FILE2]);
+	      printf_output ("=======\n");
+	      leading_dot |= dotlines (b, mapping[FILE2]);
 	    }
-	  fprintf (outputfile, ">>>>>>> %s\n", file2);
-	  undotlines (outputfile, leading_dot,
+	  printf_output (">>>>>>> %s\n", file2);
+	  undotlines (leading_dot,
 		      D_HIGHLINE (b, mapping[FILE0]) + 2,
 		      (D_NUMLINES (b, mapping[FILE1])
 		       + D_NUMLINES (b, mapping[FILE2]) + 1));
@@ -1514,17 +1618,17 @@ output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
 
 	  /* Mark start of conflict.  */
 
-	  fprintf (outputfile, "%da\n<<<<<<< %s\n",
-		   D_LOWLINE (b, mapping[FILE0]) - 1,
-		   type == DIFF_ALL ? file0 : file1);
+	  printf_output ("%da\n<<<<<<< %s\n",
+			 D_LOWLINE (b, mapping[FILE0]) - 1,
+			 type == DIFF_ALL ? file0 : file1);
 	  leading_dot = 0;
 	  if (type == DIFF_2ND)
 	    {
 	      /* Prepend lines from FILE1.  */
-	      leading_dot = dotlines (outputfile, b, mapping[FILE1]);
-	      fprintf (outputfile, "=======\n");
+	      leading_dot = dotlines (b, mapping[FILE1]);
+	      printf_output ("=======\n");
 	    }
-	  undotlines (outputfile, leading_dot,
+	  undotlines (leading_dot,
 		      D_LOWLINE (b, mapping[FILE0]) + 1,
 		      D_NUMLINES (b, mapping[FILE1]));
 	}
@@ -1532,12 +1636,11 @@ output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
 	/* Write out a delete */
 	{
 	  if (D_NUMLINES (b, mapping[FILE0]) == 1)
-	    fprintf (outputfile, "%dd\n",
-		     D_LOWLINE (b, mapping[FILE0]));
+	    printf_output ("%dd\n", D_LOWLINE (b, mapping[FILE0]));
 	  else
-	    fprintf (outputfile, "%d,%dd\n",
-		     D_LOWLINE (b, mapping[FILE0]),
-		     D_HIGHLINE (b, mapping[FILE0]));
+	    printf_output ("%d,%dd\n",
+			   D_LOWLINE (b, mapping[FILE0]),
+			   D_HIGHLINE (b, mapping[FILE0]));
 	}
       else
 	/* Write out an add or change */
@@ -1545,33 +1648,32 @@ output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
 	  switch (D_NUMLINES (b, mapping[FILE0]))
 	    {
 	    case 0:
-	      fprintf (outputfile, "%da\n",
-		       D_HIGHLINE (b, mapping[FILE0]));
+	      printf_output ("%da\n", D_HIGHLINE (b, mapping[FILE0]));
 	      break;
 	    case 1:
-	      fprintf (outputfile, "%dc\n",
-		       D_HIGHLINE (b, mapping[FILE0]));
+	      printf_output ("%dc\n", D_HIGHLINE (b, mapping[FILE0]));
 	      break;
 	    default:
-	      fprintf (outputfile, "%d,%dc\n",
-		       D_LOWLINE (b, mapping[FILE0]),
-		       D_HIGHLINE (b, mapping[FILE0]));
+	      printf_output ("%d,%dc\n",
+			     D_LOWLINE (b, mapping[FILE0]),
+			     D_HIGHLINE (b, mapping[FILE0]));
 	      break;
 	    }
 
-	  undotlines (outputfile, dotlines (outputfile, b, mapping[FILE2]),
+	  undotlines (dotlines (b, mapping[FILE2]),
 		      D_LOWLINE (b, mapping[FILE0]),
 		      D_NUMLINES (b, mapping[FILE2]));
 	}
     }
-  if (finalwrite) fprintf (outputfile, "w\nq\n");
+  if (finalwrite) printf_output ("w\nq\n");
   return conflicts_found;
 }
 
 /*
- * Read from INFILE and output to OUTPUTFILE a set of diff3_ blocks DIFF
- * as a merged file.  This acts like 'ed file0 <[output_diff3_edscript]',
- * except that it works even for binary data or incomplete lines.
+ * Read from INFILE and output to the standard output file a set of
+ * diff3_ blocks DIFF as a merged file.  This acts like 'ed file0
+ * <[output_diff3_edscript]', except that it works even for binary
+ * data or incomplete lines.
  *
  * As before, MAPPING maps from arg list file number to diff file number,
  * REV_MAPPING is its inverse,
@@ -1581,14 +1683,15 @@ output_diff3_edscript (outputfile, diff, mapping, rev_mapping,
  */
 
 static int
-output_diff3_merge (infile, outputfile, diff, mapping, rev_mapping,
+output_diff3_merge (infile, diff, mapping, rev_mapping,
 		    file0, file1, file2)
-     FILE *infile, *outputfile;
+     FILE *infile;
      struct diff3_block *diff;
      int const mapping[3], rev_mapping[3];
      char const *file0, *file1, *file2;
 {
   int c, i;
+  char cc;
   int conflicts_found = 0, conflict;
   struct diff3_block *b;
   int linesread = 0;
@@ -1627,7 +1730,8 @@ output_diff3_merge (infile, outputfile, diff, mapping, rev_mapping,
 		diff3_perror_with_exit ("input file");
 	      else if (feof (infile))
 		diff3_fatal ("input file shrank");
-	    putc (c, outputfile);
+	    cc = c;
+	    write_output (&cc, 1);
 	  }
 	while (c != '\n');
 
@@ -1638,37 +1742,37 @@ output_diff3_merge (infile, outputfile, diff, mapping, rev_mapping,
 	  if (type == DIFF_ALL)
 	    {
 	      /* Put in lines from FILE0 with bracket.  */
-	      fprintf (outputfile, "<<<<<<< %s\n", file0);
+	      printf_output ("<<<<<<< %s\n", file0);
 	      for (i = 0;
 		   i < D_NUMLINES (b, mapping[FILE0]);
 		   i++)
-		fwrite (D_RELNUM (b, mapping[FILE0], i), sizeof (char),
-			D_RELLEN (b, mapping[FILE0], i), outputfile);
+		write_output (D_RELNUM (b, mapping[FILE0], i),
+			      D_RELLEN (b, mapping[FILE0], i));
 	    }
 
 	  if (show_2nd)
 	    {
 	      /* Put in lines from FILE1 with bracket.  */
-	      fprintf (outputfile, format_2nd, file1);
+	      printf_output (format_2nd, file1);
 	      for (i = 0;
 		   i < D_NUMLINES (b, mapping[FILE1]);
 		   i++)
-		fwrite (D_RELNUM (b, mapping[FILE1], i), sizeof (char),
-			D_RELLEN (b, mapping[FILE1], i), outputfile);
+		write_output (D_RELNUM (b, mapping[FILE1], i),
+			      D_RELLEN (b, mapping[FILE1], i));
 	    }
 
-	  fprintf (outputfile, "=======\n");
+	  printf_output ("=======\n");
 	}
 
       /* Put in lines from FILE2.  */
       for (i = 0;
 	   i < D_NUMLINES (b, mapping[FILE2]);
 	   i++)
-	fwrite (D_RELNUM (b, mapping[FILE2], i), sizeof (char),
-		D_RELLEN (b, mapping[FILE2], i), outputfile);
+	write_output (D_RELNUM (b, mapping[FILE2], i),
+		      D_RELLEN (b, mapping[FILE2], i));
 
       if (conflict)
-	fprintf (outputfile, ">>>>>>> %s\n", file2);
+	printf_output (">>>>>>> %s\n", file2);
 
       /* Skip I lines in file 0.  */
       i = D_NUMLINES (b, FILE0);
@@ -1687,7 +1791,10 @@ output_diff3_merge (infile, outputfile, diff, mapping, rev_mapping,
     }
   /* Copy rest of common file.  */
   while ((c = getc (infile)) != EOF || !(ferror (infile) | feof (infile)))
-    putc (c, outputfile);
+    {
+      cc = c;
+      write_output (&cc, 1);
+    }
   return conflicts_found;
 }
 
@@ -1726,7 +1833,7 @@ static void
 diff3_fatal (string)
      char const *string;
 {
-  fprintf (stderr, "%s: %s\n", diff_program_name, string);
+  diff_error ("%s", string, 0);
   DIFF3_ABORT (2);
 }
 
@@ -1734,10 +1841,7 @@ static void
 diff3_perror_with_exit (string)
      char const *string;
 {
-  int e = errno;
-  fprintf (stderr, "%s: ", diff_program_name);
-  errno = e;
-  perror (string);
+  perror_with_name (string);
   DIFF3_ABORT (2);
 }
 
@@ -1757,4 +1861,43 @@ initialize_main (argcp, argvp)
   finalwrite = 0;
   merge = 0;
   diff_program_name = (*argvp)[0];
+  outfile = NULL;
+}
+
+static void
+free_diff_blocks(p)
+    struct diff_block *p;
+{
+  register struct diff_block *next;
+
+  while (p)
+    {
+      next = p->next;
+      if (p->lines[0]) free(p->lines[0]);
+      if (p->lines[1]) free(p->lines[1]);
+      if (p->lengths[0]) free(p->lengths[0]);
+      if (p->lengths[1]) free(p->lengths[1]);
+      free(p);
+      p = next;
+    }
+}
+
+static void
+free_diff3_blocks(p)
+    struct diff3_block *p;
+{
+  register struct diff3_block *next;
+
+  while (p)
+    {
+      next = p->next;
+      if (p->lines[0]) free(p->lines[0]);
+      if (p->lines[1]) free(p->lines[1]);
+      if (p->lines[2]) free(p->lines[2]);
+      if (p->lengths[0]) free(p->lengths[0]);
+      if (p->lengths[1]) free(p->lengths[1]);
+      if (p->lengths[2]) free(p->lengths[2]);
+      free(p);
+      p = next;
+    }
 }
