@@ -386,30 +386,6 @@ rename_file (from, to)
 	error (1, errno, "cannot rename file %s to %s", from, to);
 }
 
-/*
- * link a file, if possible.  Warning: the Windows NT version of this
- * function just copies the file, so only use this function in ways
- * that can deal with either a link or a copy.
- */
-int
-link_file (from, to)
-    const char *from;
-    const char *to;
-{
-    if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> link(%s,%s)\n",
-			(server_active) ? 'S' : ' ', from, to);
-#else
-	(void) fprintf (stderr, "-> link(%s,%s)\n", from, to);
-#endif
-    if (noexec)
-	return (0);
-
-    copy_file (from, to);
-    return 0;
-}
-
 /*
  * unlink a file, if possible.
  */
