@@ -1,17 +1,17 @@
 /*
  * Copyright (c) 1992, Brian Berliner and Jeff Polk
  * Copyright (c) 1989-1992, Brian Berliner
- * 
+ *
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
- * 
+ *
  * Commit Files
- * 
+ *
  * "commit" commits the present version to the RCS repository, AFTER
  * having done a test on conflicts.
  *
  * The call is: cvs commit [options] files...
- * 
+ *
  */
 
 #include <assert.h>
@@ -29,7 +29,7 @@ static int check_filesdoneproc PROTO ((void *callerdat, int err,
 				       char *repos, char *update_dir,
 				       List *entries));
 static int checkaddfile PROTO((char *file, char *repository, char *tag,
-			       char *options, RCSNode **rcsnode)); 
+			       char *options, RCSNode **rcsnode));
 static Dtype commit_direntproc PROTO ((void *callerdat, char *dir,
 				       char *repos, char *update_dir,
 				       List *entries));
@@ -49,7 +49,6 @@ static int precommit_list_proc PROTO((Node * p, void *closure));
 static int precommit_proc PROTO((char *repository, char *filter));
 static int remove_file PROTO ((struct file_info *finfo, char *tag,
 			       char *message));
-static void fix_rcs_modes PROTO((char *rcs, char *user));
 static void fixaddfile PROTO((char *file, char *repository));
 static void fixbranch PROTO((RCSNode *, char *branch));
 static void unlockrcs PROTO((RCSNode *rcs));
@@ -74,12 +73,13 @@ static int force_ci = 0;
 static int got_message;
 static int run_module_prog = 1;
 static int aflag;
-static char *tag;
+static char *saved_tag;
 static char *write_dirtag;
 static int write_dirnonbranch;
 static char *logfile;
 static List *mulist;
-static char *message;
+static List *saved_ulist;
+static char *saved_message;
 static time_t last_register_time;
 
 static const char *const commit_usage[] =
@@ -247,7 +247,7 @@ find_fileproc (callerdat, finfo)
     xfinfo.repository = NULL;
     xfinfo.rcs = NULL;
 
-    vers = Version_TS (&xfinfo, NULL, tag, NULL, 0, 0);
+    vers = Version_TS (&xfinfo, NULL, saved_tag, NULL, 0, 0);
     if (vers->ts_user == NULL
 	&& vers->vn_user != NULL
 	&& vers->vn_user[0] == '-')
@@ -339,7 +339,12 @@ commit (argc, argv)
     /* FIXME: Shouldn't this check be much more closely related to the
        readonly user stuff (CVSROOT/readers, &c).  That is, why should
        root be able to "cvs init", "cvs import", &c, but not "cvs ci"?  */
-    if (geteuid () == (uid_t) 0)
+    if (geteuid () == (uid_t) 0
+#  ifdef CLIENT_SUPPORT
+	/* Who we are on the client side doesn't affect logging.  */
+	&& !client_active
+#  endif
+	)
     {
 	struct passwd *pw;
 
@@ -364,18 +369,18 @@ commit (argc, argv)
 #else
 		use_editor = 0;
 #endif
-		if (message)
+		if (saved_message)
 		{
-		    free (message);
-		    message = NULL;
+		    free (saved_message);
+		    saved_message = NULL;
 		}
 
-		message = xstrdup(optarg);
+		saved_message = xstrdup(optarg);
 		break;
 	    case 'r':
-		if (tag)
-		    free (tag);
-		tag = xstrdup (optarg);
+		if (saved_tag)
+		    free (saved_tag);
+		saved_tag = xstrdup (optarg);
 		break;
 	    case 'l':
 		local = 1;
@@ -405,12 +410,12 @@ commit (argc, argv)
     argv += optind;
 
     /* numeric specified revision means we ignore sticky tags... */
-    if (tag && isdigit (*tag))
+    if (saved_tag && isdigit ((unsigned char) *saved_tag))
     {
 	aflag = 1;
 	/* strip trailing dots */
-	while (tag[strlen (tag) - 1] == '.')
-	    tag[strlen (tag) - 1] = '\0';
+	while (saved_tag[strlen (saved_tag) - 1] == '.')
+	    saved_tag[strlen (saved_tag) - 1] = '\0';
     }
 
     /* some checks related to the "-F logfile" option */
@@ -419,7 +424,7 @@ commit (argc, argv)
 	int n, logfd;
 	struct stat statbuf;
 
-	if (message)
+	if (saved_message)
 	    error (1, 0, "cannot specify both a message and a log file");
 
 	/* FIXME: Why is this binary?  Needs more investigation.  */
@@ -429,21 +434,20 @@ commit (argc, argv)
 	if (fstat(logfd, &statbuf) < 0)
 	    error (1, errno, "cannot find size of log file %s", logfile);
 
-	message = xmalloc (statbuf.st_size + 1);
+	saved_message = xmalloc (statbuf.st_size + 1);
 
 	/* FIXME: Should keep reading until EOF, rather than assuming the
 	   first read gets the whole thing.  */
-	if ((n = read (logfd, message, statbuf.st_size + 1)) < 0)
+	if ((n = read (logfd, saved_message, statbuf.st_size + 1)) < 0)
 	    error (1, errno, "cannot read log message from %s", logfile);
 
 	(void) close (logfd);
-	message[n] = '\0';
+	saved_message[n] = '\0';
     }
 
 #ifdef CLIENT_SUPPORT
-    if (client_active) 
+    if (client_active)
     {
-	int err;
 	struct find_data find_args;
 
 	ign_setup ();
@@ -458,7 +462,7 @@ commit (argc, argv)
 	   I haven't really thought about it much.
 	   Anyway, I suspect that setting it unnecessarily only causes
 	   a little unneeded network traffic.  */
-	find_args.force = force_ci || tag != NULL;
+	find_args.force = force_ci || saved_tag != NULL;
 
 	err = start_recursion (find_fileproc, find_filesdoneproc,
 			       find_dirent_proc, (DIRLEAVEPROC) NULL,
@@ -500,17 +504,17 @@ commit (argc, argv)
 	 * The protocol is designed this way.  This is a feature.
 	 */
 	if (use_editor)
-	    do_editor (".", &message, (char *)NULL, find_args.ulist);
+	    do_editor (".", &saved_message, (char *)NULL, find_args.ulist);
 
 	/* Run the user-defined script to verify/check information in
 	 *the log message
 	 */
-	do_verify (message, (char *)NULL);  	
+	do_verify (saved_message, (char *)NULL);
 
 	/* We always send some sort of message, even if empty.  */
 	/* FIXME: is that true?  There seems to be some code in do_editor
 	   which can leave the message NULL.  */
-	option_with_arg ("-m", message);
+	option_with_arg ("-m", saved_message);
 
 	/* OK, now process all the questionable files we have been saving
 	   up.  */
@@ -559,14 +563,7 @@ commit (argc, argv)
 	    send_arg("-f");
 	if (!run_module_prog)
 	    send_arg("-n");
-	option_with_arg ("-r", tag);
-
-	/* Sending only the names of the files which were modified, added,
-	   or removed means that the server will only do an up-to-date
-	   check on those files.  This is different from local CVS and
-	   previous versions of client/server CVS, but it probably is a Good
-	   Thing, or at least Not Such A Bad Thing.  */
-	send_file_names (find_args.argc, find_args.argv, 0);
+	option_with_arg ("-r", saved_tag);
 
 	/* FIXME: This whole find_args.force/SEND_FORCE business is a
 	   kludge.  It would seem to be a server bug that we have to
@@ -580,9 +577,16 @@ commit (argc, argv)
 	send_files (find_args.argc, find_args.argv, local, 0,
 		    find_args.force ? SEND_FORCE : 0);
 
+	/* Sending only the names of the files which were modified, added,
+	   or removed means that the server will only do an up-to-date
+	   check on those files.  This is different from local CVS and
+	   previous versions of client/server CVS, but it probably is a Good
+	   Thing, or at least Not Such A Bad Thing.  */
+	send_file_names (find_args.argc, find_args.argv, 0);
+
 	send_to_server ("ci\012", 0);
 	err = get_responses_and_close ();
-	if (err != 0 && use_editor && message != NULL)
+	if (err != 0 && use_editor && saved_message != NULL)
 	{
 	    /* If there was an error, don't nuke the user's carefully
 	       constructed prose.  This is something of a kludge; a better
@@ -600,7 +604,8 @@ commit (argc, argv)
 	    fp = CVS_FOPEN (fname, "w+");
 	    if (fp == NULL)
 		error (1, 0, "cannot create temporary file %s", fname);
-	    if (fwrite (message, 1, strlen (message), fp) != strlen (message))
+	    if (fwrite (saved_message, 1, strlen (saved_message), fp)
+		!= strlen (saved_message))
 		error (1, errno, "cannot write temporary file %s", fname);
 	    if (fclose (fp) < 0)
 		error (0, errno, "cannot close temporary file %s", fname);
@@ -610,12 +615,12 @@ commit (argc, argv)
     }
 #endif
 
-    if (tag != NULL)
-	tag_check_valid (tag, argc, argv, local, aflag, "");
+    if (saved_tag != NULL)
+	tag_check_valid (saved_tag, argc, argv, local, aflag, "");
 
     /* XXX - this is not the perfect check for this */
     if (argc <= 0)
-	write_dirtag = tag;
+	write_dirtag = saved_tag;
 
     wrap_setup ();
 
@@ -651,17 +656,6 @@ commit (argc, argv)
 	error (1, 0, "correct above errors first!");
     }
 
-#ifdef PRESERVE_PERMISSIONS_SUPPORT
-    if (preserve_perms)
-    {
-	/* hardlist now includes a complete index of the files
-	   to be committed, indexed by inode.  For each inode,
-	   compile a list of the files that are linked to it,
-	   and save this list in each file's hardlink_info node. */
-	(void) walklist (hardlist, cache_hardlinks_proc, NULL);
-    }
-#endif
-
     /*
      * Run the recursion processor to commit the files
      */
@@ -682,9 +676,10 @@ commit (argc, argv)
     {
 	time_t now;
 
-	(void) time (&now);
-	if (now == last_register_time) 
+	for (;;)
 	{
+	    (void) time (&now);
+	    if (now != last_register_time) break;
 	    sleep (1);			/* to avoid time-stamp races */
 	}
     }
@@ -713,10 +708,10 @@ classify_file_internal (finfo, vers)
     noexec = quiet = really_quiet = 1;
 
     /* handle specified numeric revision specially */
-    if (tag && isdigit (*tag))
+    if (saved_tag && isdigit ((unsigned char) *saved_tag))
     {
 	/* If the tag is for the trunk, make sure we're at the head */
-	if (numdots (tag) < 2)
+	if (numdots (saved_tag) < 2)
 	{
 	    status = Classify_File (finfo, (char *) NULL, (char *) NULL,
 				    (char *) NULL, 1, aflag, vers, 0);
@@ -726,7 +721,7 @@ classify_file_internal (finfo, vers)
 		Ctype xstatus;
 
 		freevers_ts (vers);
-		xstatus = Classify_File (finfo, tag, (char *) NULL,
+		xstatus = Classify_File (finfo, saved_tag, (char *) NULL,
 					 (char *) NULL, 1, aflag, vers, 0);
 		if (xstatus == T_REMOVE_ENTRY)
 		    status = T_MODIFIED;
@@ -744,7 +739,7 @@ classify_file_internal (finfo, vers)
 	     * The revision is off the main trunk; make sure we're
 	     * up-to-date with the head of the specified branch.
 	     */
-	    xtag = xstrdup (tag);
+	    xtag = xstrdup (saved_tag);
 	    if ((numdots (xtag) & 1) != 0)
 	    {
 		cp = strrchr (xtag, '.');
@@ -765,12 +760,12 @@ classify_file_internal (finfo, vers)
 	    }
 	    /* now, muck with vers to make the tag correct */
 	    free ((*vers)->tag);
-	    (*vers)->tag = xstrdup (tag);
+	    (*vers)->tag = xstrdup (saved_tag);
 	    free (xtag);
 	}
     }
     else
-	status = Classify_File (finfo, tag, (char *) NULL, (char *) NULL,
+	status = Classify_File (finfo, saved_tag, (char *) NULL, (char *) NULL,
 				1, 0, vers, 0);
     noexec = save_noexec;
     quiet = save_quiet;
@@ -796,7 +791,20 @@ check_fileproc (callerdat, finfo)
     Vers_TS *vers;
     struct commit_info *ci;
     struct logfile_info *li;
-    
+
+    size_t cvsroot_len = strlen (CVSroot_directory);
+
+    if (strncmp (finfo->repository, CVSroot_directory, cvsroot_len) == 0
+	&& ISDIRSEP (finfo->repository[cvsroot_len])
+	&& strncmp (finfo->repository + cvsroot_len + 1,
+		    CVSROOTADM,
+		    sizeof (CVSROOTADM) - 1) == 0
+	&& ISDIRSEP (finfo->repository[cvsroot_len + sizeof (CVSROOTADM)])
+	&& strcmp (finfo->repository + cvsroot_len + sizeof (CVSROOTADM) + 1,
+		   CVSNULLREPOS) == 0
+	)
+	error (1, 0, "cannot check in to %s", finfo->repository);
+
     status = classify_file_internal (finfo, &vers);
 
     /*
@@ -835,7 +843,7 @@ check_fileproc (callerdat, finfo)
 	     *    allow the commit if timestamp is identical or if we find
 	     *    an RCS_MERGE_PAT in the file.
 	     */
-	    if (!tag || !isdigit (*tag))
+	    if (!saved_tag || !isdigit ((unsigned char) *saved_tag))
 	    {
 		if (vers->date)
 		{
@@ -907,7 +915,9 @@ warning: file `%s' seems to still contain conflict indicators",
 		}
 	    }
 
-	    if (status == T_REMOVED && vers->tag && isdigit (*vers->tag))
+	    if (status == T_REMOVED
+		&& vers->tag
+		&& isdigit ((unsigned char) *vers->tag))
 	    {
 		/* Remove also tries to forbid this, but we should check
 		   here.  I'm only _sure_ about somewhat obscure cases
@@ -946,7 +956,7 @@ warning: file `%s' seems to still contain conflict indicators",
 		    }
 		    free (rcs);
 		}
-		if (vers->tag && isdigit (*vers->tag) &&
+		if (vers->tag && isdigit ((unsigned char) *vers->tag) &&
 		    numdots (vers->tag) > 1)
 		{
 		    error (0, 0,
@@ -1006,7 +1016,7 @@ warning: file `%s' seems to still contain conflict indicators",
 	    ci = (struct commit_info *) xmalloc (sizeof (struct commit_info));
 	    ci->status = status;
 	    if (vers->tag)
-		if (isdigit (*vers->tag))
+		if (isdigit ((unsigned char) *vers->tag))
 		    ci->rev = xstrdup (vers->tag);
 		else
 		    ci->rev = RCS_whatbranch (finfo->rcs, vers->tag);
@@ -1047,7 +1057,6 @@ warning: file `%s' seems to still contain conflict indicators",
 		    hlinfo = (struct hardlink_info *)
 			xmalloc (sizeof (struct hardlink_info));
 		    hlinfo->status = status;
-		    hlinfo->links = NULL;
 		    linkp->data = (char *) hlinfo;
 		}
 	    }
@@ -1114,7 +1123,6 @@ precommit_list_proc (p, closure)
 /*
  * Callback proc for pre-commit checking
  */
-static List *ulist;
 static int
 precommit_proc (repository, filter)
     char *repository;
@@ -1124,10 +1132,10 @@ precommit_proc (repository, filter)
     if (isabsolute (filter))
     {
     	char *s, *cp;
-	
+
 	s = xstrdup (filter);
 	for (cp = s; *cp; cp++)
-	    if (isspace (*cp))
+	    if (isspace ((unsigned char) *cp))
 	    {
 		*cp = '\0';
 		break;
@@ -1143,7 +1151,7 @@ precommit_proc (repository, filter)
 
     run_setup (filter);
     run_arg (repository);
-    (void) walklist (ulist, precommit_list_proc, NULL);
+    (void) walklist (saved_ulist, precommit_list_proc, NULL);
     return (run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL|RUN_REALLY));
 }
 
@@ -1165,12 +1173,12 @@ check_filesdoneproc (callerdat, err, repos, update_dir, entries)
     /* find the update list for this dir */
     p = findnode (mulist, update_dir);
     if (p != NULL)
-	ulist = ((struct master_lists *) p->data)->ulist;
+	saved_ulist = ((struct master_lists *) p->data)->ulist;
     else
-	ulist = (List *) NULL;
+	saved_ulist = (List *) NULL;
 
     /* skip the checks if there's nothing to do */
-    if (ulist == NULL || ulist->list->next == ulist->list)
+    if (saved_ulist == NULL || saved_ulist->list->next == saved_ulist->list)
 	return (err);
 
     /* run any pre-commit checks */
@@ -1238,8 +1246,9 @@ commit_fileproc (callerdat, finfo)
     {
 	got_message = 1;
 	if (use_editor)
-	    do_editor (finfo->update_dir, &message, finfo->repository, ulist);
-	do_verify (message, finfo->repository);  
+	    do_editor (finfo->update_dir, &saved_message,
+		       finfo->repository, ulist);
+	do_verify (saved_message, finfo->repository);
     }
 
     p = findnode (cilist, finfo->file);
@@ -1273,13 +1282,17 @@ commit_fileproc (callerdat, finfo)
 	   Since the branch test was done in check_fileproc for
 	   modified files, we need to stub it in again here. */
 
-	if (ci->tag)
+	if (ci->tag
+
+	    /* If numeric, it is on the trunk; check_fileproc enforced
+	       this.  */
+	    && !isdigit ((unsigned char) ci->tag[0]))
 	{
 	    if (finfo->rcs == NULL)
 		error (1, 0, "internal error: no parsed RCS file");
 	    ci->rev = RCS_whatbranch (finfo->rcs, ci->tag);
 	    err = Checkin ('A', finfo, finfo->rcs->path, ci->rev,
-			   ci->tag, ci->options, message);
+			   ci->tag, ci->options, saved_message);
 	    if (err != 0)
 	    {
 		unlockrcs (finfo->rcs);
@@ -1319,7 +1332,7 @@ commit_fileproc (callerdat, finfo)
     {
 	err = Checkin ('M', finfo,
 		       finfo->rcs->path, ci->rev, ci->tag,
-		       ci->options, message);
+		       ci->options, saved_message);
 
 	(void) time (&last_register_time);
 
@@ -1331,7 +1344,7 @@ commit_fileproc (callerdat, finfo)
     }
     else if (ci->status == T_REMOVED)
     {
-	err = remove_file (finfo, ci->tag, message);
+	err = remove_file (finfo, ci->tag, saved_message);
 #ifdef SERVER_SUPPORT
 	if (server_active) {
 	    server_scratch_entry_only ();
@@ -1371,7 +1384,7 @@ out:
            been removed from the archive, which is not the behavior we
            want for our commitlog messages; we want the old version
            number and then "NONE." */
-	
+
 	if (ci->status != T_REMOVED)
 	{
 	    p = findnode (ulist, finfo->file);
@@ -1379,7 +1392,7 @@ out:
 	    {
 		Vers_TS *vers;
 		struct logfile_info *li;
-	    
+
 		(void) classify_file_internal (finfo, &vers);
 		li = (struct logfile_info *) p->data;
 		li->rev_new = xstrdup (vers->vn_rcs);
@@ -1415,7 +1428,7 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
     got_message = 0;
 
 
-    Update_Logfile (repository, message, (FILE *) 0, ulist);
+    Update_Logfile (repository, saved_message, (FILE *) 0, ulist);
 
     /* Build the administrative files if necessary.  */
     {
@@ -1423,11 +1436,17 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 
 	if (strncmp (CVSroot_directory, repository,
 		     strlen (CVSroot_directory)) != 0)
-	    error (0, 0, "internal error: repository (%s) doesn't begin with root (%s)", repository, CVSroot_directory);
+	    error (0, 0,
+		 "internal error: repository (%s) doesn't begin with root (%s)",
+		   repository, CVSroot_directory);
 	p = repository + strlen (CVSroot_directory);
 	if (*p == '/')
 	    ++p;
-	if (strcmp ("CVSROOT", p) == 0)
+	if (strcmp ("CVSROOT", p) == 0
+	    /* Check for subdirectories because people may want to create
+	       subdirectories and list files therein in checkoutlist.  */
+	    || strncmp ("CVSROOT/", p, strlen ("CVSROOT/")) == 0
+	    )
 	{
 	    /* "Database" might a little bit grandiose and/or vague,
 	       but "checked-out copies of administrative files, unless
@@ -1435,11 +1454,21 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 	       case modules.{pag,dir,db}" is verbose and excessively
 	       focused on how the database is implemented.  */
 
+	    /* mkmodules requires the absolute name of the CVSROOT directory.
+	       Remove anything after the `CVSROOT' component -- this is
+	       necessary when committing in a subdirectory of CVSROOT.  */
+	    char *admin_dir = xstrdup (repository);
+	    int cvsrootlen = strlen ("CVSROOT");
+	    assert (admin_dir[p - repository + cvsrootlen] == '\0'
+		    || admin_dir[p - repository + cvsrootlen] == '/');
+	    admin_dir[p - repository + cvsrootlen] = '\0';
+
 	    cvs_output (program_name, 0);
 	    cvs_output (" ", 1);
 	    cvs_output (command_name, 0);
 	    cvs_output (": Rebuilding administrative file database\n", 0);
-	    mkmodules (repository);
+	    mkmodules (admin_dir);
+	    free (admin_dir);
 	}
     }
 
@@ -1452,7 +1481,7 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 	    char *line;
 	    int line_length;
 	    size_t line_chars_allocated;
-	    char *repository;
+	    char *repos;
 
 	    line = NULL;
 	    line_chars_allocated = 0;
@@ -1462,9 +1491,9 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 		/* Remove any trailing newline.  */
 		if (line[line_length - 1] == '\n')
 		    line[--line_length] = '\0';
-		repository = Name_Repository ((char *) NULL, update_dir);
+		repos = Name_Repository ((char *) NULL, update_dir);
 		run_setup (line);
-		run_arg (repository);
+		run_arg (repos);
 		cvs_output (program_name, 0);
 		cvs_output (" ", 1);
 		cvs_output (command_name, 0);
@@ -1472,7 +1501,7 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 		run_print (stdout);
 		cvs_output ("'\n", 0);
 		(void) run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL);
-		free (repository);
+		free (repos);
 	    }
 	    else
 	    {
@@ -1529,8 +1558,8 @@ commit_direntproc (callerdat, dir, repos, update_dir, entries)
     real_repos = Name_Repository (dir, update_dir);
     got_message = 1;
     if (use_editor)
-	do_editor (update_dir, &message, real_repos, ulist);
-    do_verify (message, real_repos);  
+	do_editor (update_dir, &saved_message, real_repos, ulist);
+    do_verify (saved_message, real_repos);
     free (real_repos);
     return (R_PROCESS);
 }
@@ -1591,16 +1620,16 @@ findmaxrev (p, closure)
  * XXX - if removing a ,v file that is a relative symbolic link to
  * another ,v file, we probably should add a ".." component to the
  * link to keep it relative after we move it into the attic.
- */
+
+   Return value is 0 on success, or >0 on error (in which case we have
+   printed an error message).  */
 static int
 remove_file (finfo, tag, message)
     struct file_info *finfo;
     char *tag;
     char *message;
 {
-    mode_t omask;
     int retcode;
-    char *tmp;
 
     int branch;
     int lockflag;
@@ -1622,7 +1651,7 @@ remove_file (finfo, tag, message)
     if (tag && !(branch = RCS_nodeisbranch (finfo->rcs, tag)))
     {
 	/* a symbolic tag is specified; just remove the tag from the file */
-	if ((retcode = RCS_deltag (finfo->rcs, tag)) != 0) 
+	if ((retcode = RCS_deltag (finfo->rcs, tag)) != 0)
 	{
 	    if (!quiet)
 		error (0, retcode == -1 ? errno : 0,
@@ -1655,7 +1684,7 @@ remove_file (finfo, tag, message)
 	    error (0, 0, "cannot find branch \"%s\".", tag);
 	    return (1);
 	}
-	
+
 	branchname = RCS_getbranch (finfo->rcs, rev, 1);
 	if (branchname == NULL)
 	{
@@ -1676,12 +1705,12 @@ remove_file (finfo, tag, message)
         /* Get current head revision of file. */
 	prev_rev = RCS_head (finfo->rcs);
     }
-    
+
     /* if removing without a tag or a branch, then make sure the default
        branch is the trunk. */
     if (!tag && !branch)
     {
-        if (RCS_setbranch (finfo->rcs, NULL) != 0) 
+        if (RCS_setbranch (finfo->rcs, NULL) != 0)
 	{
 	    error (0, 0, "cannot change branch to default for %s",
 		   finfo->fullname);
@@ -1690,16 +1719,6 @@ remove_file (finfo, tag, message)
 	RCS_rewrite (finfo->rcs, NULL, NULL);
     }
 
-#ifdef SERVER_SUPPORT
-    if (server_active) {
-	/* If this is the server, there will be a file sitting in the
-	   temp directory which is the kludgy way in which server.c
-	   tells time_stamp that the file is no longer around.  Remove
-	   it so we can create temp files with that name (ignore errors).  */
-	unlink_file (finfo->file);
-    }
-#endif
-
     /* check something out.  Generally this is the head.  If we have a
        particular rev, then name it.  */
     retcode = RCS_checkout (finfo->rcs, finfo->file, rev ? corev : NULL,
@@ -1736,33 +1755,9 @@ remove_file (finfo, tag, message)
     if (rev != NULL)
 	free (rev);
 
-    old_path = finfo->rcs->path;
+    old_path = xstrdup (finfo->rcs->path);
     if (!branch)
-    {
-	/* this was the head; really move it into the Attic */
-	tmp = xmalloc(strlen(finfo->repository) + 
-		      sizeof('/') +
-		      sizeof(CVSATTIC) +
-		      sizeof('/') +
-		      strlen(finfo->file) +
-		      sizeof(RCSEXT) + 1);
-	(void) sprintf (tmp, "%s/%s", finfo->repository, CVSATTIC);
-	omask = umask (cvsumask);
-	(void) CVS_MKDIR (tmp, 0777);
-	(void) umask (omask);
-	(void) sprintf (tmp, "%s/%s/%s%s", finfo->repository, CVSATTIC, finfo->file, RCSEXT);
-	
-	if (strcmp (finfo->rcs->path, tmp) != 0
-	    && CVS_RENAME (finfo->rcs->path, tmp) == -1
-	    && (isreadable (finfo->rcs->path) || !isreadable (tmp)))
-	{
-	    free(tmp);
-	    return (1);
-	}
-	/* The old value of finfo->rcs->path is in old_path, and is
-           freed below.  */
-	finfo->rcs->path = tmp;
-    }
+	RCS_setattic (finfo->rcs, 1);
 
     /* Print message that file was removed. */
     cvs_output (old_path, 0);
@@ -1773,8 +1768,7 @@ remove_file (finfo, tag, message)
     cvs_output ("\ndone\n", 0);
     free(prev_rev);
 
-    if (old_path != finfo->rcs->path)
-	free (old_path);
+    free (old_path);
 
     Scratch_Entry (finfo->entries, finfo->file);
     return (0);
@@ -1794,13 +1788,15 @@ finaladd (finfo, rev, tag, options)
     char *rcs;
 
     rcs = locate_rcs (finfo->file, finfo->repository);
-    ret = Checkin ('A', finfo, rcs, rev, tag, options, message);
+    ret = Checkin ('A', finfo, rcs, rev, tag, options, saved_message);
     if (ret == 0)
     {
 	char *tmp = xmalloc (strlen (finfo->file) + sizeof (CVSADM)
 			     + sizeof (CVSEXT_LOG) + 10);
 	(void) sprintf (tmp, "%s/%s%s", CVSADM, finfo->file, CVSEXT_LOG);
-	(void) unlink_file (tmp);
+	if (unlink_file (tmp) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", tmp);
 	free (tmp);
     }
     else
@@ -1844,7 +1840,10 @@ fixaddfile (file, repository)
     save_really_quiet = really_quiet;
     really_quiet = 1;
     if ((rcsfile = RCS_parsercsfile (rcs)) == NULL)
-	(void) unlink_file (rcs);
+    {
+	if (unlink_file (rcs) < 0)
+	    error (0, errno, "cannot remove %s", rcs);
+    }
     else
 	freercsnode (&rcsfile);
     really_quiet = save_really_quiet;
@@ -1891,8 +1890,20 @@ checkaddfile (file, repository, tag, options, rcsnode)
     int newfile = 0;
     RCSNode *rcsfile = NULL;
     int retval;
+    int adding_on_branch;
 
-    if (tag)
+    /* Callers expect to be able to use either "" or NULL to mean the
+       default keyword expansion.  */
+    if (options != NULL && options[0] == '\0')
+	options = NULL;
+    if (options != NULL)
+	assert (options[0] == '-' && options[1] == 'k');
+
+    /* If numeric, it is on the trunk; check_fileproc enforced
+       this.  */
+    adding_on_branch = tag != NULL && !isdigit ((unsigned char) tag[0]);
+
+    if (adding_on_branch)
     {
 	rcs = xmalloc (strlen (repository) + strlen (file)
 		       + sizeof (RCSEXT) + sizeof (CVSATTIC) + 10);
@@ -1915,6 +1926,7 @@ checkaddfile (file, repository, tag, options, rcsnode)
     {
 	/* file has existed in the past.  Prepare to resurrect. */
 	char *rev;
+	char *oldexpand;
 
 	if ((rcsfile = *rcsnode) == NULL)
 	{
@@ -1923,41 +1935,38 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	    goto out;
 	}
 
-	if (tag == NULL)
+	oldexpand = RCS_getexpand (rcsfile);
+	if ((oldexpand != NULL
+	     && options != NULL
+	     && strcmp (options + 2, oldexpand) != 0)
+	    || (oldexpand == NULL && options != NULL))
 	{
-	    char *oldfile;
+	    /* We tell the user about this, because it means that the
+	       old revisions will no longer retrieve the way that they
+	       used to.  */
+	    error (0, 0, "changing keyword expansion mode to %s", options);
+	    RCS_setexpand (rcsfile, options + 2);
+	}
 
-	    /* we are adding on the trunk, so move the file out of the
-	       Attic. */
-	    oldfile = xstrdup (rcs);
-	    sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
-	    
-	    if (strcmp (oldfile, rcs) == 0)
+	if (!adding_on_branch)
+	{
+	    /* We are adding on the trunk, so move the file out of the
+	       Attic.  */
+	    if (!(rcsfile->flags & INATTIC))
 	    {
 		error (0, 0, "internal error: confused about attic for %s",
-		       oldfile);
-	    out1:
-		free (oldfile);
+		       rcsfile->path);
 		retval = 1;
 		goto out;
 	    }
-	    if (CVS_RENAME (oldfile, rcs) != 0)
+
+	    sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
+
+	    if (RCS_setattic (rcsfile, 0))
 	    {
-		error (0, errno, "failed to move `%s' out of the attic",
-		       oldfile);
-		goto out1;
+		retval = 1;
+		goto out;
 	    }
-	    if (isreadable (oldfile)
-		|| !isreadable (rcs))
-	    {
-		error (0, 0, "\
-internal error: `%s' didn't move out of the attic",
-		       oldfile);
-		goto out1;
-	    }
-	    free (oldfile);
-	    free (rcsfile->path);
-	    rcsfile->path = xstrdup (rcs);
 	}
 
 	rev = RCS_getversion (rcsfile, tag, NULL, 1, (int *) NULL);
@@ -2010,7 +2019,7 @@ internal error: `%s' didn't move out of the attic",
 	}
 
 	/* Set RCS keyword expansion options.  */
-	if (options && options[0] == '-' && options[1] == 'k')
+	if (options != NULL)
 	    opt = options + 2;
 	else
 	    opt = NULL;
@@ -2039,7 +2048,7 @@ internal error: `%s' didn't move out of the attic",
 
     /* when adding a file for the first time, and using a tag, we need
        to create a dead revision on the trunk.  */
-    if (tag && newfile)
+    if (adding_on_branch && newfile)
     {
 	char *tmp;
 	FILE *fp;
@@ -2101,7 +2110,7 @@ internal error: `%s' didn't move out of the attic",
 	}
     }
 
-    if (tag != NULL)
+    if (adding_on_branch)
     {
 	/* when adding with a tag, we need to stub a branch, if it
 	   doesn't already exist.  */
@@ -2154,7 +2163,7 @@ internal error: `%s' didn't move out of the attic",
 		retval = 1;
 		goto out;
 	    }
-	} 
+	}
 
 	if (rcsnode && *rcsnode != rcsfile)
 	{
@@ -2165,13 +2174,22 @@ internal error: `%s' didn't move out of the attic",
 
     fileattr_newfile (file);
 
-    /* I don't think fix_rcs_modes is needed any more.  In the
-       add_rcs_file case, the algorithms used by add_rcs_file and
-       fix_rcs_modes are the same, so there is no need to go through
-       it all twice.  In the other cases, I think we want to just
-       preserve the mode that the file had before we started.  That is
-       a behavior change, but I would think a desirable one.  */
-    fix_rcs_modes (rcs, file);
+    /* At this point, we used to set the file mode of the RCS file
+       based on the mode of the file in the working directory.  If we
+       are creating the RCS file for the first time, add_rcs_file does
+       this already.  If we are re-adding the file, then perhaps it is
+       consistent to preserve the old file mode, just as we preserve
+       the old keyword expansion mode.
+
+       If we decide that we should change the modes, then we can't do
+       it here anyhow.  At this point, the RCS file may be owned by
+       somebody else, so a chmod will fail.  We need to instead do the
+       chmod after rewriting it.
+
+       FIXME: In general, I think the file mode (and the keyword
+       expansion mode) should be associated with a particular revision
+       of the file, so that it is possible to have different revisions
+       of a file have different modes.  */
 
     retval = 0;
 
@@ -2200,12 +2218,13 @@ lock_RCS (user, rcs, rev, repository)
      * For a specified, numeric revision of the form "1" or "1.1", (or when
      * no revision is specified ""), definitely move the branch to the trunk
      * before locking the RCS file.
-     * 
+     *
      * The assumption is that if there is more than one revision on the trunk,
      * the head points to the trunk, not a branch... and as such, it's not
      * necessary to move the head in this case.
      */
-    if (rev == NULL || (rev && isdigit (*rev) && numdots (rev) < 2))
+    if (rev == NULL
+	|| (rev && isdigit ((unsigned char) *rev) && numdots (rev) < 2))
     {
 	branch = xstrdup (rcs->branch);
 	if (branch != NULL)
@@ -2262,69 +2281,6 @@ lock_RCS (user, rcs, rev, repository)
     return (1);
 }
 
-/* Called when "add"ing files to the RCS respository.  It doesn't seem to
-   be possible to get RCS to use the right mode, so we change it after
-   the fact.  TODO: now that RCS has been librarified, we have the power
-   to change this. */
-
-static void
-fix_rcs_modes (rcs, user)
-    char *rcs;
-    char *user;
-{
-    struct stat sb;
-    mode_t rcs_mode;
-
-#ifdef PRESERVE_PERMISSIONS_SUPPORT
-    /* Do ye nothing to the modes on a symbolic link. */
-    if (preserve_perms && islink (user))
-	return;
-#endif
-
-    if (CVS_STAT (user, &sb) < 0)
-    {
-	/* FIXME: Should be ->fullname.  */
-	error (0, errno, "warning: cannot stat %s", user);
-	return;
-    }
-
-    /* Now we compute the new mode.
-
-       TODO: decide whether this whole thing can/should be skipped
-       when `preserve_perms' is set.  Almost certainly so. -twp
-
-       The algorithm that we use is:
-
-       Write permission is always off (this is what RCS and CVS have always
-       done).
-
-       If S_IRUSR is on (user read), then the read permission of
-       the RCS file will be on.  It would seem that if this is off,
-       then other users can't do "cvs update" and such, so perhaps this
-       should be hardcoded to being on (it is a strange case, though--the
-       case in which a user file doesn't have user read permission on).
-
-       If S_IXUSR is on (user execute), then set execute permission
-       on the RCS file.  This allows other users who check out the file
-       to get the right setting for whether a shell script (for example)
-       has the executable bit set.
-
-       The result of that calculation is modified by CVSUMASK.  The
-       reason, of course, that the read and execute settings take the
-       user bit and copy it to all three bits (user, group, other), is
-       that it should be CVSUMASK, not the umask of individual users,
-       which is the sole determiner of modes in the repository.  */
-
-    rcs_mode = 0;
-    if (sb.st_mode & S_IRUSR)
-	rcs_mode |= S_IRUSR | S_IRGRP | S_IROTH;
-    if (sb.st_mode & S_IXUSR)
-	rcs_mode |= S_IXUSR | S_IXGRP | S_IXOTH;
-    rcs_mode &= ~cvsumask;
-    if (chmod (rcs, rcs_mode) < 0)
-	error (0, errno, "warning: cannot change mode of %s", rcs);
-}
-
 /*
  * free an UPDATE node's data
  */
