@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Entries file to Files file
  * 
@@ -27,7 +27,12 @@ static char *entfilename;		/* for error messages */
 /*
  * Construct an Entnode
  */
-Entnode *
+static Entnode *Entnode_Create PROTO ((enum ent_type, const char *,
+				       const char *, const char *,
+				       const char *, const char *,
+				       const char *, const char *));
+
+static Entnode *
 Entnode_Create(type, user, vn, ts, options, tag, date, ts_conflict)
     enum ent_type type;
     const char *user;
@@ -57,7 +62,9 @@ Entnode_Create(type, user, vn, ts, options, tag, date, ts_conflict)
 /*
  * Destruct an Entnode
  */
-void
+static void Entnode_Destroy PROTO ((Entnode *));
+
+static void
 Entnode_Destroy (ent)
     Entnode *ent;
 {
@@ -110,7 +117,26 @@ write_entries (list)
 
     /* open the new one and walk the list writing entries */
     entfilename = CVSADM_ENTBAK;
-    entfile = open_file (entfilename, "w+");
+    entfile = CVS_FOPEN (entfilename, "w+");
+    if (entfile == NULL)
+    {
+	/* Make this a warning, not an error.  For example, one user might
+	   have checked out a working directory which, for whatever reason,
+	   contains an Entries.Log file.  A second user, without write access
+	   to that working directory, might want to do a "cvs log".  The
+	   problem rewriting Entries shouldn't affect the ability of "cvs log"
+	   to work, although the warning is probably a good idea so that
+	   whether Entries gets rewritten is not an inexplicable process.  */
+	/* FIXME: should be including update_dir in message.  */
+	error (0, errno, "cannot rewrite %s", entfilename);
+
+	/* Now just return.  We leave the Entries.Log file around.  As far
+	   as I know, there is never any data lying around in 'list' that
+	   is not in Entries.Log at this time (if there is an error writing
+	   Entries.Log that is a separate problem).  */
+	return;
+    }
+
     (void) walklist (list, write_ent_proc, (void *) &sawdir);
     if (! sawdir)
     {
@@ -131,7 +157,9 @@ write_entries (list)
     rename_file (entfilename, CVSADM_ENT);
 
     /* now, remove the log file */
-    unlink_file (CVSADM_ENTLOG);
+    if (unlink_file (CVSADM_ENTLOG) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot remove %s", CVSADM_ENTLOG);
 }
 
 /*
@@ -145,12 +173,8 @@ Scratch_Entry (list, fname)
     Node *node;
 
     if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> Scratch_Entry(%s)\n",
-			(server_active) ? 'S' : ' ', fname);
-#else
-	(void) fprintf (stderr, "-> Scratch_Entry(%s)\n", fname);
-#endif
+	(void) fprintf (stderr, "%s-> Scratch_Entry(%s)\n",
+			CLIENT_SERVER_STR, fname);
 
     /* hashlookup to see if it is there */
     if ((node = findnode_fn (list, fname)) != NULL)
@@ -205,18 +229,11 @@ Register (list, fname, vn, ts, options, tag, date, ts_conflict)
 
     if (trace)
     {
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> Register(%s, %s, %s%s%s, %s, %s %s)\n",
-			(server_active) ? 'S' : ' ',
+	(void) fprintf (stderr, "%s-> Register(%s, %s, %s%s%s, %s, %s %s)\n",
+			CLIENT_SERVER_STR,
 			fname, vn, ts ? ts : "",
 			ts_conflict ? "+" : "", ts_conflict ? ts_conflict : "",
 			options, tag ? tag : "", date ? date : "");
-#else
-	(void) fprintf (stderr, "-> Register(%s, %s, %s%s%s, %s, %s %s)\n",
-			fname, vn, ts ? ts : "",
-			ts_conflict ? "+" : "", ts_conflict ? ts_conflict : "",
-			options, tag ? tag : "", date ? date : "");
-#endif
     }
 
     entnode = Entnode_Create (ENT_FILE, fname, vn, ts, options, tag, date,
@@ -226,7 +243,15 @@ Register (list, fname, vn, ts, options, tag, date, ts_conflict)
     if (!noexec)
     {
 	entfilename = CVSADM_ENTLOG;
-	entfile = open_file (entfilename, "a");
+	entfile = CVS_FOPEN (entfilename, "a");
+
+	if (entfile == NULL)
+	{
+	    /* Warning, not error, as in write_entries.  */
+	    /* FIXME-update-dir: should be including update_dir in message.  */
+	    error (0, errno, "cannot open %s", entfilename);
+	    return;
+	}
 
 	if (fprintf (entfile, "A ") < 0)
 	    error (1, errno, "cannot write %s", entfilename);
@@ -255,6 +280,9 @@ freesdt (p)
     free ((char *) sdtp);
 }
 
+/* Return the next real Entries line.  On end of file, returns NULL.
+   On error, prints an error message and returns NULL.  */
+
 static Entnode *
 fgetentent(fpin, cmd, sawdir)
     FILE *fpin;
@@ -268,12 +296,13 @@ fgetentent(fpin, cmd, sawdir)
     enum ent_type type;
     char *l, *user, *vn, *ts, *options;
     char *tag_or_date, *tag, *date, *ts_conflict;
+    int line_length;
 
     line = NULL;
     line_chars_allocated = 0;
 
     ent = NULL;
-    while (getline (&line, &line_chars_allocated, fpin) > 0)
+    while ((line_length = getline (&line, &line_chars_allocated, fpin)) > 0)
     {
 	l = line;
 
@@ -368,6 +397,9 @@ fgetentent(fpin, cmd, sawdir)
 	break;
     }
 
+    if (line_length < 0 && !feof (fpin))
+	error (0, errno, "cannot read entries file");
+
     free (line);
     return ent;
 }
@@ -417,17 +449,21 @@ fputentent(fp, p)
 }
 
 
-/*
- * Read the entries file into a list, hashing on the file name.
- */
+/* Read the entries file into a list, hashing on the file name.
+
+   UPDATE_DIR is the name of the current directory, for use in error
+   messages, or NULL if not known (that is, noone has gotten around
+   to updating the caller to pass in the information).  */
 List *
-Entries_Open (aflag)
+Entries_Open (aflag, update_dir)
     int aflag;
+    char *update_dir;
 {
     List *entries;
     struct stickydirtag *sdtp = NULL;
     Entnode *ent;
     char *dirtag, *dirdate;
+    int dirnonbranch;
     int do_rewrite = 0;
     FILE *fpin;
     int sawdir;
@@ -439,7 +475,7 @@ Entries_Open (aflag)
      * Parse the CVS/Tag file, to get any default tag/date settings. Use
      * list-private storage to tuck them away for Version_TS().
      */
-    ParseTag (&dirtag, &dirdate);
+    ParseTag (&dirtag, &dirdate, &dirnonbranch);
     if (aflag || dirtag || dirdate)
     {
 	sdtp = (struct stickydirtag *) xmalloc (sizeof (*sdtp));
@@ -447,6 +483,7 @@ Entries_Open (aflag)
 	sdtp->aflag = aflag;
 	sdtp->tag = xstrdup (dirtag);
 	sdtp->date = xstrdup (dirdate);
+	sdtp->nonbranch = dirnonbranch;
 
 	/* feed it into the list-private area */
 	entries->list->data = (char *) sdtp;
@@ -457,7 +494,11 @@ Entries_Open (aflag)
 
     fpin = CVS_FOPEN (CVSADM_ENT, "r");
     if (fpin == NULL)
+    {
+	if (update_dir != NULL)
+	    error (0, 0, "in directory %s:", update_dir);
 	error (0, errno, "cannot open %s for reading", CVSADM_ENT);
+    }
     else
     {
 	while ((ent = fgetentent (fpin, (char *) NULL, &sawdir)) != NULL) 
@@ -465,7 +506,9 @@ Entries_Open (aflag)
 	    (void) AddEntryNode (entries, ent);
 	}
 
-	fclose (fpin);
+	if (fclose (fpin) < 0)
+	    /* FIXME-update-dir: should include update_dir in message.  */
+	    error (0, errno, "cannot close %s", CVSADM_ENT);
     }
 
     fpin = CVS_FOPEN (CVSADM_ENTLOG, "r");
@@ -493,7 +536,9 @@ Entries_Open (aflag)
 	    }
 	}
 	do_rewrite = 1;
-	fclose (fpin);
+	if (fclose (fpin) < 0)
+	    /* FIXME-update-dir: should include update_dir in message.  */
+	    error (0, errno, "cannot close %s", CVSADM_ENTLOG);
     }
 
     /* Update the list private data to indicate whether subdirectory
@@ -590,17 +635,23 @@ AddEntryNode (list, entdata)
  * Write out/Clear the CVS/Tag file.
  */
 void
-WriteTag (dir, tag, date)
+WriteTag (dir, tag, date, nonbranch, update_dir, repository)
     char *dir;
     char *tag;
     char *date;
+    int nonbranch;
+    char *update_dir;
+    char *repository;
 {
     FILE *fout;
-    char tmp[PATH_MAX];
+    char *tmp;
 
     if (noexec)
 	return;
 
+    tmp = xmalloc ((dir ? strlen (dir) : 0)
+		   + sizeof (CVSADM_TAG)
+		   + 10);
     if (dir == NULL)
 	(void) strcpy (tmp, CVSADM_TAG);
     else
@@ -611,8 +662,16 @@ WriteTag (dir, tag, date)
 	fout = open_file (tmp, "w+");
 	if (tag)
 	{
-	    if (fprintf (fout, "T%s\n", tag) < 0)
-		error (1, errno, "write to %s failed", tmp);
+	    if (nonbranch)
+	    {
+		if (fprintf (fout, "N%s\n", tag) < 0)
+		    error (1, errno, "write to %s failed", tmp);
+	    }
+	    else
+	    {
+		if (fprintf (fout, "T%s\n", tag) < 0)
+		    error (1, errno, "write to %s failed", tmp);
+	    }
 	}
 	else
 	{
@@ -625,15 +684,35 @@ WriteTag (dir, tag, date)
     else
 	if (unlink_file (tmp) < 0 && ! existence_error (errno))
 	    error (1, errno, "cannot remove %s", tmp);
+    free (tmp);
+#ifdef SERVER_SUPPORT
+    if (server_active)
+	server_set_sticky (update_dir, repository, tag, date, nonbranch);
+#endif
 }
 
-/*
- * Parse the CVS/Tag file for the current directory.
- */
+/* Parse the CVS/Tag file for the current directory.
+
+   If it contains a date, sets *DATEP to the date in a newly malloc'd
+   string, *TAGP to NULL, and *NONBRANCHP to an unspecified value.
+
+   If it contains a branch tag, sets *TAGP to the tag in a newly
+   malloc'd string, *NONBRANCHP to 0, and *DATEP to NULL.
+
+   If it contains a nonbranch tag, sets *TAGP to the tag in a newly
+   malloc'd string, *NONBRANCHP to 1, and *DATEP to NULL.
+
+   If it does not exist, or contains something unrecognized by this
+   version of CVS, set *DATEP and *TAGP to NULL and *NONBRANCHP to
+   an unspecified value.
+
+   If there is an error, print an error message, set *DATEP and *TAGP
+   to NULL, and return.  */
 void
-ParseTag (tagp, datep)
+ParseTag (tagp, datep, nonbranchp)
     char **tagp;
     char **datep;
+    int *nonbranchp;
 {
     FILE *fp;
 
@@ -641,6 +720,11 @@ ParseTag (tagp, datep)
 	*tagp = (char *) NULL;
     if (datep)
 	*datep = (char *) NULL;
+    /* Always store a value here, even in the 'D' case where the value
+       is unspecified.  Shuts up tools which check for references to
+       uninitialized memory.  */
+    if (nonbranchp != NULL)
+	*nonbranchp = 0;
     fp = CVS_FOPEN (CVSADM_TAG, "r");
     if (fp)
     {
@@ -650,20 +734,54 @@ ParseTag (tagp, datep)
 
 	line = NULL;
 	line_chars_allocated = 0;
-	  
+
 	if ((line_length = getline (&line, &line_chars_allocated, fp)) > 0)
 	{
 	    /* Remove any trailing newline.  */
 	    if (line[line_length - 1] == '\n')
 	        line[--line_length] = '\0';
-	    if (*line == 'T' && tagp)
-		*tagp = xstrdup (line + 1);
-	    else if (*line == 'D' && datep)
-		*datep = xstrdup (line + 1);
+	    switch (*line)
+	    {
+		case 'T':
+		    if (tagp != NULL)
+			*tagp = xstrdup (line + 1);
+		    break;
+		case 'D':
+		    if (datep != NULL)
+			*datep = xstrdup (line + 1);
+		    break;
+		case 'N':
+		    if (tagp != NULL)
+			*tagp = xstrdup (line + 1);
+		    if (nonbranchp != NULL)
+			*nonbranchp = 1;
+		    break;
+		default:
+		    /* Silently ignore it; it may have been
+		       written by a future version of CVS which extends the
+		       syntax.  */
+		    break;
+	    }
 	}
-	(void) fclose (fp);
+
+	if (line_length < 0)
+	{
+	    /* FIXME-update-dir: should include update_dir in messages.  */
+	    if (feof (fp))
+		error (0, 0, "cannot read %s: end of file", CVSADM_TAG);
+	    else
+		error (0, errno, "cannot read %s", CVSADM_TAG);
+	}
+
+	if (fclose (fp) < 0)
+	    /* FIXME-update-dir: should include update_dir in message.  */
+	    error (0, errno, "cannot close %s", CVSADM_TAG);
+
 	free (line);
     }
+    else if (!existence_error (errno))
+	/* FIXME-update-dir: should include update_dir in message.  */
+	error (0, errno, "cannot open %s", CVSADM_TAG);
 }
 
 /*
@@ -686,10 +804,26 @@ Subdirs_Known (entries)
 	FILE *fp;
 
 	sdtp->subdirs = 1;
-	/* Create Entries.Log so that Entries_Close will do something.  */
-	fp = open_file (CVSADM_ENTLOG, "a");
-	if (fclose (fp) == EOF)
-	    error (1, errno, "cannot close %s", CVSADM_ENTLOG);
+	if (!noexec)
+	{
+	    /* Create Entries.Log so that Entries_Close will do something.  */
+	    fp = CVS_FOPEN (CVSADM_ENTLOG, "a");
+	    if (fp == NULL)
+	    {
+		int save_errno = errno;
+
+		/* As in subdir_record, just silently skip the whole thing
+		   if there is no CVSADM directory.  */
+		if (! isdir (CVSADM))
+		    return;
+		error (1, save_errno, "cannot open %s", entfilename);
+	    }
+	    else
+	    {
+		if (fclose (fp) == EOF)
+		    error (1, errno, "cannot close %s", CVSADM_ENTLOG);
+	    }
+	}
     }
 }
 
@@ -821,3 +955,230 @@ Subdir_Deregister (entries, parent, dir)
 	    delnode (p);
     }
 }
+
+
+
+/* OK, the following base_* code tracks the revisions of the files in
+   CVS/Base.  We do this in a file CVS/Baserev.  Separate from
+   CVS/Entries because it needs to go in separate data structures
+   anyway (the name in Entries must be unique), so this seemed
+   cleaner.  The business of rewriting the whole file in
+   base_deregister and base_register is the kind of thing we used to
+   do for Entries and which turned out to be slow, which is why there
+   is now the Entries.Log machinery.  So maybe from that point of
+   view it is a mistake to do this separately from Entries, I dunno.
+
+   We also need something analogous for:
+
+   1. CVS/Template (so we can update the Template file automagically
+   without the user needing to check out a new working directory).
+   Updating would probably print a message (that part might be
+   optional, although probably it should be visible because not all
+   cvs commands would make the update happen and so it is a
+   user-visible behavior).  Constructing version number for template
+   is a bit hairy (base it on the timestamp on the server?  Or see if
+   the template is in checkoutlist and if yes use its versioning and
+   if no don't version it?)....
+
+   2.  cvsignore (need to keep a copy in the working directory to do
+   "cvs release" on the client side; see comment at src/release.c
+   (release).  Would also allow us to stop needing Questionable.  */
+
+enum base_walk {
+    /* Set the revision for FILE to *REV.  */
+    BASE_REGISTER,
+    /* Get the revision for FILE and put it in a newly malloc'd string
+       in *REV, or put NULL if not mentioned.  */
+    BASE_GET,
+    /* Remove FILE.  */
+    BASE_DEREGISTER
+};
+
+static void base_walk PROTO ((enum base_walk, struct file_info *, char **));
+
+/* Read through the lines in CVS/Baserev, taking the actions as documented
+   for CODE.  */
+
+static void
+base_walk (code, finfo, rev)
+    enum base_walk code;
+    struct file_info *finfo;
+    char **rev;
+{
+    FILE *fp;
+    char *line;
+    size_t line_allocated;
+    FILE *newf;
+    char *baserev_fullname;
+    char *baserevtmp_fullname;
+
+    line = NULL;
+    line_allocated = 0;
+    newf = NULL;
+
+    /* First compute the fullnames for the error messages.  This
+       computation probably should be broken out into a separate function,
+       as recurse.c does it too and places like Entries_Open should be
+       doing it.  */
+    baserev_fullname = xmalloc (sizeof (CVSADM_BASEREV)
+				+ strlen (finfo->update_dir)
+				+ 2);
+    baserev_fullname[0] = '\0';
+    baserevtmp_fullname = xmalloc (sizeof (CVSADM_BASEREVTMP)
+				   + strlen (finfo->update_dir)
+				   + 2);
+    baserevtmp_fullname[0] = '\0';
+    if (finfo->update_dir[0] != '\0')
+    {
+	strcat (baserev_fullname, finfo->update_dir);
+	strcat (baserev_fullname, "/");
+	strcat (baserevtmp_fullname, finfo->update_dir);
+	strcat (baserevtmp_fullname, "/");
+    }
+    strcat (baserev_fullname, CVSADM_BASEREV);
+    strcat (baserevtmp_fullname, CVSADM_BASEREVTMP);
+
+    fp = CVS_FOPEN (CVSADM_BASEREV, "r");
+    if (fp == NULL)
+    {
+	if (!existence_error (errno))
+	{
+	    error (0, errno, "cannot open %s for reading", baserev_fullname);
+	    goto out;
+	}
+    }
+
+    switch (code)
+    {
+	case BASE_REGISTER:
+	case BASE_DEREGISTER:
+	    newf = CVS_FOPEN (CVSADM_BASEREVTMP, "w");
+	    if (newf == NULL)
+	    {
+		error (0, errno, "cannot open %s for writing",
+		       baserevtmp_fullname);
+		goto out;
+	    }
+	    break;
+	case BASE_GET:
+	    *rev = NULL;
+	    break;
+    }
+
+    if (fp != NULL)
+    {
+	while (getline (&line, &line_allocated, fp) >= 0)
+	{
+	    char *linefile;
+	    char *p;
+	    char *linerev;
+
+	    if (line[0] != 'B')
+		/* Ignore, for future expansion.  */
+		continue;
+
+	    linefile = line + 1;
+	    p = strchr (linefile, '/');
+	    if (p == NULL)
+		/* Syntax error, ignore.  */
+		continue;
+	    linerev = p + 1;
+	    p = strchr (linerev, '/');
+	    if (p == NULL)
+		continue;
+
+	    linerev[-1] = '\0';
+	    if (fncmp (linefile, finfo->file) == 0)
+	    {
+		switch (code)
+		{
+		case BASE_REGISTER:
+		case BASE_DEREGISTER:
+		    /* Don't copy over the old entry, we don't want it.  */
+		    break;
+		case BASE_GET:
+		    *p = '\0';
+		    *rev = xstrdup (linerev);
+		    *p = '/';
+		    goto got_it;
+		}
+	    }
+	    else
+	    {
+		linerev[-1] = '/';
+		switch (code)
+		{
+		case BASE_REGISTER:
+		case BASE_DEREGISTER:
+		    if (fprintf (newf, "%s\n", line) < 0)
+			error (0, errno, "error writing %s",
+			       baserevtmp_fullname);
+		    break;
+		case BASE_GET:
+		    break;
+		}
+	    }
+	}
+	if (ferror (fp))
+	    error (0, errno, "cannot read %s", baserev_fullname);
+    }
+ got_it:
+
+    if (code == BASE_REGISTER)
+    {
+	if (fprintf (newf, "B%s/%s/\n", finfo->file, *rev) < 0)
+	    error (0, errno, "error writing %s",
+		   baserevtmp_fullname);
+    }
+
+ out:
+
+    if (line != NULL)
+	free (line);
+
+    if (fp != NULL)
+    {
+	if (fclose (fp) < 0)
+	    error (0, errno, "cannot close %s", baserev_fullname);
+    }
+    if (newf != NULL)
+    {
+	if (fclose (newf) < 0)
+	    error (0, errno, "cannot close %s", baserevtmp_fullname);
+	rename_file (CVSADM_BASEREVTMP, CVSADM_BASEREV);
+    }
+
+    free (baserev_fullname);
+    free (baserevtmp_fullname);
+}
+
+/* Return, in a newly malloc'd string, the revision for FILE in CVS/Baserev,
+   or NULL if not listed.  */
+
+char *
+base_get (finfo)
+    struct file_info *finfo;
+{
+    char *rev;
+    base_walk (BASE_GET, finfo, &rev);
+    return rev;
+}
+
+/* Set the revision for FILE to REV.  */
+
+void
+base_register (finfo, rev)
+    struct file_info *finfo;
+    char *rev;
+{
+    base_walk (BASE_REGISTER, finfo, &rev);
+}
+
+/* Remove FILE.  */
+
+void
+base_deregister (finfo)
+    struct file_info *finfo;
+{
+    base_walk (BASE_DEREGISTER, finfo, NULL);
+}
