@@ -43,7 +43,7 @@ variable_set (nameval)
     Node *node;
 
     p = nameval;
-    while (isalnum (*p) || *p == '_')
+    while (isalnum ((unsigned char) *p) || *p == '_')
 	++p;
     if (*p != '=')
 	error (1, 0, "illegal character in user variable name in %s", nameval);
@@ -95,40 +95,82 @@ expand_path (name, file, line)
 {
     char *s;
     char *d;
-    /* FIXME: arbitrary limit.  */
-    char  mybuf[PATH_MAX];
-    char  buf[PATH_MAX];
+
+    char *mybuf = NULL;
+    size_t mybuf_size = 0;
+    char *buf = NULL;
+    size_t buf_size = 0;
+
+    size_t doff;
+
     char *result;
+
+    /* Sorry this routine is so ugly; it is a head-on collision
+       between the `traditional' unix *d++ style and the need to
+       dynamically allocate.  It would be much cleaner (and probably
+       faster, not that this is a bottleneck for CVS) with more use of
+       strcpy & friends, but I haven't taken the effort to rewrite it
+       thusly.  */
+
+    /* First copy from NAME to MYBUF, expanding $<foo> as we go.  */
     s = name;
     d = mybuf;
+    doff = d - mybuf;
+    expand_string (&mybuf, &mybuf_size, doff + 1);
+    d = mybuf + doff;
     while ((*d++ = *s))
+    {
 	if (*s++ == '$')
 	{
 	    char *p = d;
 	    char *e;
 	    int flag = (*s == '{');
 
+	    doff = d - mybuf;
+	    expand_string (&mybuf, &mybuf_size, doff + 1);
+	    d = mybuf + doff;
 	    for (; (*d++ = *s); s++)
+	    {
 		if (flag
 		    ? *s =='}'
-		    : isalnum (*s) == 0 && *s != '_')
+		    : isalnum ((unsigned char) *s) == 0 && *s != '_')
 		    break;
-	    *--d = 0;
+		doff = d - mybuf;
+		expand_string (&mybuf, &mybuf_size, doff + 1);
+		d = mybuf + doff;
+	    }
+	    *--d = '\0';
 	    e = expand_variable (&p[flag], file, line);
 
 	    if (e)
 	    {
+		doff = d - mybuf;
+		expand_string (&mybuf, &mybuf_size, doff + 1);
+		d = mybuf + doff;
 		for (d = &p[-1]; (*d++ = *e++);)
-		    ;
+		{
+		    doff = d - mybuf;
+		    expand_string (&mybuf, &mybuf_size, doff + 1);
+		    d = mybuf + doff;
+		}
 		--d;
 		if (flag && *s)
 		    s++;
 	    }
 	    else
 		/* expand_variable has already printed an error message.  */
-		return NULL;
+		goto error_exit;
 	}
-    *d = 0;
+	doff = d - mybuf;
+	expand_string (&mybuf, &mybuf_size, doff + 1);
+	d = mybuf + doff;
+    }
+    doff = d - mybuf;
+    expand_string (&mybuf, &mybuf_size, doff + 1);
+    d = mybuf + doff;
+    *d = '\0';
+
+    /* Then copy from MYBUF to BUF, expanding ~.  */
     s = mybuf;
     d = buf;
     /* If you don't want ~username ~/ to be expanded simply remove
@@ -142,6 +184,19 @@ expand_path (name, file, line)
 	    t = get_homedir ();
 	else
 	{
+#ifdef GETPWNAM_MISSING
+	    for (; *p!='/' && *p; p++)
+		;
+	    *p = 0;
+	    if (line != 0)
+		error (0, 0,
+		       "%s:%d:tilde expansion not supported on this system",
+		       file, line);
+	    else
+		error (0, 0, "%s:tilde expansion not supported on this system",
+		       file);
+	    return NULL;
+#else
 	    struct passwd *ps;
 	    for (; *p!='/' && *p; p++)
 		;
@@ -157,9 +212,20 @@ expand_path (name, file, line)
 		return NULL;
 	    }
 	    t = ps->pw_dir;
+#endif
 	}
+	if (t == NULL)
+	    error (1, 0, "cannot find home directory");
+
+	doff = d - buf;
+	expand_string (&buf, &buf_size, doff + 1);
+	d = buf + doff;
 	while ((*d++ = *t++))
-	    ;
+	{
+	    doff = d - buf;
+	    expand_string (&buf, &buf_size, doff + 1);
+	    d = buf + doff;
+	}
 	--d;
 	if (*p == 0)
 	    *p = '/';	       /* always add / */
@@ -168,12 +234,35 @@ expand_path (name, file, line)
     else
 	--s;
 	/* Kill up to here */
+    doff = d - buf;
+    expand_string (&buf, &buf_size, doff + 1);
+    d = buf + doff;
     while ((*d++ = *s++))
-	;
-    *d=0;
-    result = xmalloc (sizeof(char) * strlen(buf)+1);
-    strcpy (result, buf);
+    {
+	doff = d - buf;
+	expand_string (&buf, &buf_size, doff + 1);
+	d = buf + doff;
+    }
+    doff = d - buf;
+    expand_string (&buf, &buf_size, doff + 1);
+    d = buf + doff;
+    *d = '\0';
+
+    /* OK, buf contains the value we want to return.  Clean up and return
+       it.  */
+    free (mybuf);
+    /* Save a little memory with xstrdup; buf will tend to allocate
+       more than it needs to.  */
+    result = xstrdup (buf);
+    free (buf);
     return result;
+
+ error_exit:
+    if (mybuf != NULL)
+	free (mybuf);
+    if (buf != NULL)
+	free (buf);
+    return NULL;
 }
 
 static char *
@@ -184,8 +273,11 @@ expand_variable (name, file, line)
 {
     if (strcmp (name, CVSROOT_ENV) == 0)
 	return CVSroot_original;
-    else if (strcmp (name, RCSBIN_ENV)  == 0)
-	return Rcsbin;
+    else if (strcmp (name, "RCSBIN") == 0)
+    {
+	error (0, 0, "RCSBIN internal variable is no longer supported");
+	return NULL;
+    }
     else if (strcmp (name, EDITOR1_ENV) == 0)
 	return Editor;
     else if (strcmp (name, EDITOR2_ENV) == 0)
@@ -194,7 +286,7 @@ expand_variable (name, file, line)
 	return Editor;
     else if (strcmp (name, "USER") == 0)
 	return getcaller ();
-    else if (isalpha (name[0]))
+    else if (isalpha ((unsigned char) name[0]))
     {
 	/* These names are reserved for future versions of CVS,
 	   so that is why it is an error.  */
