@@ -1,10 +1,8 @@
-/* $CVSid: @(#)hash.h 1.23 94/10/07 $	 */
-
 /*
  * Copyright (c) 1992, Brian Berliner and Jeff Polk
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  */
 
 /*
@@ -19,7 +17,8 @@
 enum ntype
 {
     UNKNOWN, HEADER, ENTRIES, FILES, LIST, RCSNODE,
-    RCSVERS, DIRS, UPDATE, LOCK, NDBMNODE
+    RCSVERS, DIRS, UPDATE, LOCK, NDBMNODE, FILEATTR,
+    VARIABLE, RCSFIELD
 };
 typedef enum ntype Ntype;
 
@@ -44,23 +43,17 @@ struct list
 };
 typedef struct list List;
 
-struct entnode
-{
-    char *version;
-    char *timestamp;
-    char *options;
-    char *tag;
-    char *date;
-    char *conflict;
-};
-typedef struct entnode Entnode;
-
 List *getlist PROTO((void));
 Node *findnode PROTO((List * list, const char *key));
+Node *findnode_fn PROTO((List * list, const char *key));
 Node *getnode PROTO((void));
+int insert_before PROTO((List * list, Node * marker, Node * p));
 int addnode PROTO((List * list, Node * p));
+int addnode_at_front PROTO((List * list, Node * p));
 int walklist PROTO((List * list, int (*)(Node *n, void *closure), void *closure));
+int list_isempty PROTO ((List *list));
 void dellist PROTO((List ** listp));
 void delnode PROTO((Node * p));
 void freenode PROTO((Node * p));
 void sortlist PROTO((List * list, int (*)(const Node *, const Node *)));
+int fsortcmp PROTO((const Node * p, const Node * q));
