@@ -7,13 +7,14 @@
 
 #include "cvs.h"
 #include "savecwd.h"
+#include "getline.h"
 
 #ifndef DBLKSIZ
 #define	DBLKSIZ	4096			/* since GNU ndbm doesn't define it */
 #endif
 
 static int checkout_file PROTO((char *file, char *temp));
-static void make_tempfile PROTO((char *temp));
+static char *make_tempfile PROTO((void));
 static void rename_rcsfile PROTO((char *temp, char *real));
 
 #ifndef MY_NDBM
@@ -28,37 +29,50 @@ struct admin_file {
 
    /* This is a one line description of what the file is for.  It is not
       currently used, although one wonders whether it should be, somehow.
-      If NULL, then don't process this file in mkmodules (FIXME: a bit of
+      If NULL, then don't process this file in mkmodules (FIXME?: a bit of
       a kludge; probably should replace this with a flags field).  */
    char *errormsg;
 
    /* Contents which the file should have in a new repository.  To avoid
       problems with brain-dead compilers which choke on long string constants,
       this is a pointer to an array of char * terminated by NULL--each of
-      the strings is concatenated.  */
+      the strings is concatenated.
+
+      If this field is NULL, the file is not created in a new
+      repository, but it can be added with "cvs add" (just as if one
+      had created the repository with a version of CVS which didn't
+      know about the file) and the checked-out copy will be updated
+      without having to add it to checkoutlist.  */
    const char * const *contents;
 };
 
 static const char *const loginfo_contents[] = {
-    "# The \"loginfo\" file is used to control where \"cvs commit\" log information is\n",
-    "# sent.  The first entry on a line is a regular expression which is tested\n",
-    "# against the directory that the change is being made to, relative to the\n",
-    "# $CVSROOT.  For the first match that is found, the remainder of the line is a\n",
-    "# filter program that should expect log information on its standard input\n",
+    "# The \"loginfo\" file controls where \"cvs commit\" log information\n",
+    "# is sent.  The first entry on a line is a regular expression which must match\n",
+    "# the directory that the change is being made to, relative to the\n",
+    "# $CVSROOT.  If a match is found, then the remainder of the line is a filter\n",
+    "# program that should expect log information on its standard input.\n",
     "#\n",
-    "# If the repository name does not match any of the regular expressions in the\n",
-    "# first field of this file, the \"DEFAULT\" line is used, if it is specified.\n",
+    "# If the repository name does not match any of the regular expressions in this\n",
+    "# file, the \"DEFAULT\" line is used, if it is specified.\n",
     "#\n",
-    "# If the name \"ALL\" appears as a regular expression it is always used\n",
-    "# in addition to the first matching regex or \"DEFAULT\".\n",
+    "# If the name ALL appears as a regular expression it is always used\n",
+    "# in addition to the first matching regex or DEFAULT.\n",
     "#\n",
-    "# The filter program may use one and only one \"%s\" modifier (ala printf).  If\n",
-    "# such a \"%s\" is specified in the filter program, a brief title is included\n",
-    "# (as one argument, enclosed in single quotes) showing the relative directory\n",
-    "# name and listing the modified file names.\n",
+    "# You may specify a format string as part of the\n",
+    "# filter.  The string is composed of a `%' followed\n",
+    "# by a single format character, or followed by a set of format\n",
+    "# characters surrounded by `{' and `}' as separators.  The format\n",
+    "# characters are:\n",
+    "#\n",
+    "#   s = file name\n",
+    "#   V = old version number (pre-checkin)\n",
+    "#   v = new version number (post-checkin)\n",
     "#\n",
     "# For example:\n",
-    "#DEFAULT		(echo \"\"; who am i; echo %s; date; cat) >> $CVSROOT/CVSROOT/commitlog\n",
+    "#DEFAULT (echo \"\"; id; echo %s; date; cat) >> $CVSROOT/CVSROOT/commitlog\n",
+    "# or\n",
+    "#DEFAULT (echo \"\"; id; echo %{sVv}; date; cat) >> $CVSROOT/CVSROOT/commitlog\n",
     NULL
 };
 
@@ -104,6 +118,31 @@ static const char *const editinfo_contents[] = {
     NULL
 };
 
+static const char *const verifymsg_contents[] = {
+    "# The \"verifymsg\" file is used to allow verification of logging\n",
+    "# information.  It works best when a template (as specified in the\n",
+    "# rcsinfo file) is provided for the logging procedure.  Given a\n",
+    "# template with locations for, a bug-id number, a list of people who\n",
+    "# reviewed the code before it can be checked in, and an external\n",
+    "# process to catalog the differences that were code reviewed, the\n",
+    "# following test can be applied to the code:\n",
+    "#\n",
+    "#   Making sure that the entered bug-id number is correct.\n",
+    "#   Validating that the code that was reviewed is indeed the code being\n",
+    "#       checked in (using the bug-id number or a seperate review\n",
+    "#       number to identify this particular code set.).\n",
+    "#\n",
+    "# If any of the above test failed, then the commit would be aborted.\n",
+    "#\n",
+    "# Actions such as mailing a copy of the report to each reviewer are\n",
+    "# better handled by an entry in the loginfo file.\n",
+    "#\n",
+    "# One thing that should be noted is the the ALL keyword is not\n",
+    "# supported.  There can be only one entry that matches a given\n",
+    "# repository.\n",
+    NULL
+};
+
 static const char *const commitinfo_contents[] = {
     "# The \"commitinfo\" file is used to control pre-commit checks.\n",
     "# The filter on the right is invoked with the repository and a list \n",
@@ -165,15 +204,15 @@ static const char *const checkoutlist_contents[] = {
 };
 
 static const char *const cvswrappers_contents[] = {
-    "# This file describes wrappers and other binary files to CVS.\n",
+    "# This file affects handling of files based on their names.\n",
     "#\n",
-    "# Wrappers are the concept where directories of files are to be\n",
-    "# treated as a single file.  The intended use is to wrap up a wrapper\n",
-    "# into a single tar such that the tar archive can be treated as a\n",
-    "# single binary file in CVS.\n",
+    "# The -t/-f options allow one to treat directories of files\n",
+    "# as a single file, or to transform a file in other ways on\n",
+    "# its way in and out of CVS.\n",
     "#\n",
-    "# To solve the problem effectively, it was also necessary to be able to\n",
-    "# prevent rcsmerge from merging these files.\n",
+    "# The -m option specifies whether CVS attempts to merge files.\n",
+    "#\n",
+    "# The -k option specifies keyword expansion (e.g. -kb for binary).\n",
     "#\n",
     "# Format of wrapper file ($CVSROOT/CVSROOT/cvswrappers or .cvswrappers)\n",
     "#\n",
@@ -183,10 +222,11 @@ static const char *const cvswrappers_contents[] = {
     "#  -f		from cvs filter		value: path to filter\n",
     "#  -t		to cvs filter		value: path to filter\n",
     "#  -m		update methodology	value: MERGE or COPY\n",
+    "#  -k		expansion mode		value: b, o, kkv, &c\n",
     "#\n",
     "#  and value is a single-quote delimited value.\n",
-    "#\n",
     "# For example:\n",
+    "#*.gif -k 'b'\n",
     NULL
 };
 
@@ -236,6 +276,21 @@ static const char *const modules_contents[] = {
     NULL
 };
 
+static const char *const config_contents[] = {
+    "# Set this to \"no\" if pserver shouldn't check system users/passwords\n",
+    "#SystemAuth=no\n",
+    "\n",
+    "# Set `PreservePermissions' to `yes' to save file status information\n",
+    "# in the repository.\n",
+    "#PreservePermissions=no\n",
+    "\n",
+    "# Set `TopLevelAdmin' to `yes' to create a CVS directory at the top\n",
+    "# level of the new working directory when using the `cvs checkout'\n",
+    "# command.\n",
+    "#TopLevelAdmin=no\n",
+    NULL
+};
+
 static const struct admin_file filelist[] = {
     {CVSROOTADM_LOGINFO, 
 	"no logging of 'cvs commit' messages is done without a %s file",
@@ -246,6 +301,9 @@ static const struct admin_file filelist[] = {
     {CVSROOTADM_EDITINFO,
 	"a %s file can be used to validate log messages",
 	editinfo_contents},
+    {CVSROOTADM_VERIFYMSG,
+	"a %s file can be used to validate log messages",
+	verifymsg_contents},
     {CVSROOTADM_COMMITINFO,
 	"a %s file can be used to configure 'cvs commit' checking",
 	commitinfo_contents},
@@ -268,7 +326,34 @@ static const struct admin_file filelist[] = {
 	/* modules is special-cased in mkmodules.  */
 	NULL,
 	modules_contents},
-    {NULL, NULL}
+    {CVSROOTADM_READERS,
+	"a %s file specifies read-only users",
+	NULL},
+    {CVSROOTADM_WRITERS,
+	"a %s file specifies read/write users",
+	NULL},
+
+    /* Some have suggested listing CVSROOTADM_PASSWD here too.  This
+       would mean that CVS commands which operate on the
+       CVSROOTADM_PASSWD file would transmit hashed passwords over the
+       net.  This might seem to be no big deal, as pserver normally
+       transmits cleartext passwords, but the difference is that
+       CVSROOTADM_PASSWD contains *all* passwords, not just the ones
+       currently being used.  For example, it could be too easy to
+       accidentally give someone readonly access to CVSROOTADM_PASSWD
+       (e.g. via anonymous CVS or cvsweb), and then if there are any
+       guessable passwords for read/write access (usually there will be)
+       they get read/write access.
+
+       Another worry is the implications of storing old passwords--if
+       someone used a password in the past they might be using it
+       elsewhere, using a similar password, etc, and so saving old
+       passwords, even hashed, is probably not a good idea.  */
+
+    {CVSROOTADM_CONFIG,
+	 "a %s file configures various behaviors",
+	 config_contents},
+    {NULL, NULL, NULL}
 };
 
 /* Rebuild the checked out administrative files in directory DIR.  */
@@ -277,19 +362,18 @@ mkmodules (dir)
     char *dir;
 {
     struct saved_cwd cwd;
-    /* FIXME: arbitrary limit */
-    char temp[PATH_MAX];
+    char *temp;
     char *cp, *last, *fname;
 #ifdef MY_NDBM
     DBM *db;
 #endif
     FILE *fp;
-    /* FIXME: arbitrary limit */
-    char line[512];
+    char *line = NULL;
+    size_t line_allocated = 0;
     const struct admin_file *fileptr;
 
     if (save_cwd (&cwd))
-	exit (EXIT_FAILURE);
+	error_exit ();
 
     if ( CVS_CHDIR (dir) < 0)
 	error (1, errno, "cannot chdir to %s", dir);
@@ -297,7 +381,7 @@ mkmodules (dir)
     /*
      * First, do the work necessary to update the "modules" database.
      */
-    make_tempfile (temp);
+    temp = make_tempfile ();
     switch (checkout_file (CVSROOTADM_MODULES, temp))
     {
 
@@ -313,11 +397,6 @@ mkmodules (dir)
 	    rename_rcsfile (temp, CVSROOTADM_MODULES);
 	    break;
 
-	case -1:			/* fork failed */
-	    (void) unlink_file (temp);
-	    exit (EXIT_FAILURE);
-	    /* NOTREACHED */
-
 	default:
 	    error (0, 0,
 		"'cvs checkout' is less functional without a %s file",
@@ -325,13 +404,16 @@ mkmodules (dir)
 	    break;
     }					/* switch on checkout_file() */
 
-    (void) unlink_file (temp);
+    if (unlink_file (temp) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot remove %s", temp);
+    free (temp);
 
     /* Checkout the files that need it in CVSROOT dir */
     for (fileptr = filelist; fileptr && fileptr->filename; fileptr++) {
 	if (fileptr->errormsg == NULL)
 	    continue;
-	make_tempfile (temp);
+	temp = make_tempfile ();
 	if (checkout_file (fileptr->filename, temp) == 0)
 	    rename_rcsfile (temp, fileptr->filename);
 #if 0
@@ -345,10 +427,12 @@ mkmodules (dir)
 	else if (fileptr->errormsg)
 	    error (0, 0, fileptr->errormsg, fileptr->filename);
 #endif
-	(void) unlink_file (temp);
+	if (unlink_file (temp) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", temp);
+	free (temp);
     }
 
-    /* Use 'fopen' instead of 'open_file' because we want to ignore error */
     fp = CVS_FOPEN (CVSROOTADM_CHECKOUTLIST, "r");
     if (fp)
     {
@@ -358,7 +442,7 @@ mkmodules (dir)
 	 *
 	 * comment lines begin with '#'
 	 */
-	while (fgets (line, sizeof (line), fp) != NULL)
+	while (getline (&line, &line_allocated, fp) >= 0)
 	{
 	    /* skip lines starting with # */
 	    if (line[0] == '#')
@@ -368,32 +452,51 @@ mkmodules (dir)
 		*last = '\0';			/* strip the newline */
 
 	    /* Skip leading white space. */
-	    for (fname = line; *fname && isspace(*fname); fname++)
+	    for (fname = line;
+		 *fname && isspace ((unsigned char) *fname);
+		 fname++)
 		;
 
 	    /* Find end of filename. */
-	    for (cp = fname; *cp && !isspace(*cp); cp++)
+	    for (cp = fname; *cp && !isspace ((unsigned char) *cp); cp++)
 		;
 	    *cp = '\0';
 
-	    make_tempfile (temp);
+	    temp = make_tempfile ();
 	    if (checkout_file (fname, temp) == 0)
 	    {
 		rename_rcsfile (temp, fname);
 	    }
 	    else
 	    {
-		for (cp++; cp < last && *last && isspace(*last); cp++)
+		for (cp++;
+		     cp < last && *last && isspace ((unsigned char) *last);
+		     cp++)
 		    ;
 		if (cp < last && *cp)
 		    error (0, 0, cp, fname);
 	    }
+	    if (unlink_file (temp) < 0
+		&& !existence_error (errno))
+		error (0, errno, "cannot remove %s", temp);
+	    free (temp);
 	}
-	(void) fclose (fp);
+	if (line)
+	    free (line);
+	if (ferror (fp))
+	    error (0, errno, "cannot read %s", CVSROOTADM_CHECKOUTLIST);
+	if (fclose (fp) < 0)
+	    error (0, errno, "cannot close %s", CVSROOTADM_CHECKOUTLIST);
+    }
+    else
+    {
+	/* Error from CVS_FOPEN.  */
+	if (!existence_error (errno))
+	    error (0, errno, "cannot open %s", CVSROOTADM_CHECKOUTLIST);
     }
 
     if (restore_cwd (&cwd, NULL))
-	exit (EXIT_FAILURE);
+	error_exit ();
     free_cwd (&cwd);
 
     return (0);
@@ -402,15 +505,16 @@ mkmodules (dir)
 /*
  * Yeah, I know, there are NFS race conditions here.
  */
-static void
-make_tempfile (temp)
-    char *temp;
+static char *
+make_tempfile ()
 {
     static int seed = 0;
     int fd;
+    char *temp;
 
     if (seed == 0)
 	seed = getpid ();
+    temp = xmalloc (sizeof (BAKPREFIX) + 40);
     while (1)
     {
 	(void) sprintf (temp, "%s%d", BAKPREFIX, seed++);
@@ -421,25 +525,45 @@ make_tempfile (temp)
     }
     if (close(fd) < 0)
 	error(1, errno, "cannot close temporary file %s", temp);
+    return temp;
 }
 
+/* Get a file.  If the file does not exist, return 1 silently.  If
+   there is an error, print a message and return 1 (FIXME: probably
+   not a very clean convention).  On success, return 0.  */
+
 static int
 checkout_file (file, temp)
     char *file;
     char *temp;
 {
-    char rcs[PATH_MAX];
+    char *rcs;
+    RCSNode *rcsnode;
     int retcode = 0;
 
-    (void) sprintf (rcs, "%s%s", file, RCSEXT);
+    if (noexec)
+	return 0;
+
+    rcs = xmalloc (strlen (file) + 5);
+    strcpy (rcs, file);
+    strcat (rcs, RCSEXT);
     if (!isfile (rcs))
-	return (1);
-    run_setup ("%s%s -x,v/ -q -p", Rcsbin, RCS_CO);
-    run_arg (rcs);
-    if ((retcode = run_exec (RUN_TTY, temp, RUN_TTY, RUN_NORMAL)) != 0)
     {
-	error (0, retcode == -1 ? errno : 0, "failed to check out %s file", file);
+	free (rcs);
+	return (1);
     }
+    rcsnode = RCS_parsercsfile (rcs);
+    retcode = RCS_checkout (rcsnode, NULL, NULL, NULL, NULL, temp,
+			    (RCSCHECKOUTPROC) NULL, (void *) NULL);
+    if (retcode != 0)
+    {
+	/* Probably not necessary (?); RCS_checkout already printed a
+	   message.  */
+	error (0, 0, "failed to check out %s file",
+	       file);
+    }
+    freercsnode (&rcsnode);
+    free (rcs);
     return (retcode);
 }
 
@@ -495,7 +619,7 @@ write_dbmfile (temp)
 	if (value[0] == '#')
 	    continue;			/* comment line */
 	vp = value;
-	while (*vp && isspace (*vp))
+	while (*vp && isspace ((unsigned char) *vp))
 	    vp++;
 	if (*vp == '\0')
 	    continue;			/* empty line */
@@ -506,11 +630,11 @@ write_dbmfile (temp)
 	if (!cont)
 	{
 	    key.dptr = vp;
-	    while (*vp && !isspace (*vp))
+	    while (*vp && !isspace ((unsigned char) *vp))
 		vp++;
 	    key.dsize = vp - key.dptr;
 	    *vp++ = '\0';		/* NULL terminate the key */
-	    while (*vp && isspace (*vp))
+	    while (*vp && isspace ((unsigned char) *vp))
 		vp++;			/* skip whitespace to value */
 	    if (*vp == '\0')
 	    {
@@ -527,17 +651,28 @@ write_dbmfile (temp)
 	}
     }
     dbm_close (db);
-    (void) fclose (fp);
+    if (fclose (fp) < 0)
+	error (0, errno, "cannot close %s", temp);
     if (err)
     {
+	/* I think that the size of the buffer needed here is
+	   just determined by sizeof (CVSROOTADM_MODULES), the
+	   filenames created by make_tempfile, and other things that won't
+	   overflow.  */
 	char dotdir[50], dotpag[50], dotdb[50];
 
 	(void) sprintf (dotdir, "%s.dir", temp);
 	(void) sprintf (dotpag, "%s.pag", temp);
 	(void) sprintf (dotdb, "%s.db", temp);
-	(void) unlink_file (dotdir);
-	(void) unlink_file (dotpag);
-	(void) unlink_file (dotdb);
+	if (unlink_file (dotdir) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", dotdir);
+	if (unlink_file (dotpag) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", dotpag);
+	if (unlink_file (dotdb) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", dotdb);
 	error (1, 0, "DBM creation failed; correct above errors");
     }
 }
@@ -546,10 +681,18 @@ static void
 rename_dbmfile (temp)
     char *temp;
 {
+    /* I think that the size of the buffer needed here is
+       just determined by sizeof (CVSROOTADM_MODULES), the
+       filenames created by make_tempfile, and other things that won't
+       overflow.  */
     char newdir[50], newpag[50], newdb[50];
     char dotdir[50], dotpag[50], dotdb[50];
     char bakdir[50], bakpag[50], bakdb[50];
 
+    int dir1_errno = 0, pag1_errno = 0, db1_errno = 0;
+    int dir2_errno = 0, pag2_errno = 0, db2_errno = 0;
+    int dir3_errno = 0, pag3_errno = 0, db3_errno = 0;
+
     (void) sprintf (dotdir, "%s.dir", CVSROOTADM_MODULES);
     (void) sprintf (dotpag, "%s.pag", CVSROOTADM_MODULES);
     (void) sprintf (dotdb, "%s.db", CVSROOTADM_MODULES);
@@ -567,18 +710,59 @@ rename_dbmfile (temp)
     /* don't mess with me */
     SIG_beginCrSect ();
 
-    (void) unlink_file (bakdir);	/* rm .#modules.dir .#modules.pag */
-    (void) unlink_file (bakpag);
-    (void) unlink_file (bakdb);
-    (void) CVS_RENAME (dotdir, bakdir);	/* mv modules.dir .#modules.dir */
-    (void) CVS_RENAME (dotpag, bakpag);	/* mv modules.pag .#modules.pag */
-    (void) CVS_RENAME (dotdb, bakdb);	/* mv modules.db .#modules.db */
-    (void) CVS_RENAME (newdir, dotdir);	/* mv "temp".dir modules.dir */
-    (void) CVS_RENAME (newpag, dotpag);	/* mv "temp".pag modules.pag */
-    (void) CVS_RENAME (newdb, dotdb);	/* mv "temp".db modules.db */
+    /* rm .#modules.dir .#modules.pag */
+    if (unlink_file (bakdir) < 0)
+	dir1_errno = errno;
+    if (unlink_file (bakpag) < 0)
+	pag1_errno = errno;
+    if (unlink_file (bakdb) < 0)
+	db1_errno = errno;
+
+    /* mv modules.dir .#modules.dir */
+    if (CVS_RENAME (dotdir, bakdir) < 0)
+	dir2_errno = errno;
+    /* mv modules.pag .#modules.pag */
+    if (CVS_RENAME (dotpag, bakpag) < 0)
+	pag2_errno = errno;
+    /* mv modules.db .#modules.db */
+    if (CVS_RENAME (dotdb, bakdb) < 0)
+	db2_errno = errno;
+
+    /* mv "temp".dir modules.dir */
+    if (CVS_RENAME (newdir, dotdir) < 0)
+	dir3_errno = errno;
+    /* mv "temp".pag modules.pag */
+    if (CVS_RENAME (newpag, dotpag) < 0)
+	pag3_errno = errno;
+    /* mv "temp".db modules.db */
+    if (CVS_RENAME (newdb, dotdb) < 0)
+	db3_errno = errno;
 
     /* OK -- make my day */
     SIG_endCrSect ();
+
+    /* I didn't want to call error() when we had signals blocked
+       (unnecessary?), but do it now.  */
+    if (dir1_errno && !existence_error (dir1_errno))
+	error (0, dir1_errno, "cannot remove %s", bakdir);
+    if (pag1_errno && !existence_error (pag1_errno))
+	error (0, pag1_errno, "cannot remove %s", bakpag);
+    if (db1_errno && !existence_error (db1_errno))
+	error (0, db1_errno, "cannot remove %s", bakdb);
+
+    if (dir2_errno && !existence_error (dir2_errno))
+	error (0, dir2_errno, "cannot remove %s", bakdir);
+    if (pag2_errno && !existence_error (pag2_errno))
+	error (0, pag2_errno, "cannot remove %s", bakpag);
+    if (db2_errno && !existence_error (db2_errno))
+	error (0, db2_errno, "cannot remove %s", bakdb);
+
+    if (dir3_errno && !existence_error (dir3_errno))
+	error (0, dir3_errno, "cannot remove %s", bakdir);
+    if (pag3_errno && !existence_error (pag3_errno))
+	error (0, pag3_errno, "cannot remove %s", bakpag);
+    if (db3_errno && !existence_error (db3_errno))
+	error (0, db3_errno, "cannot remove %s", bakdb);
 }
 
 #endif				/* !MY_NDBM */
@@ -588,25 +772,45 @@ rename_rcsfile (temp, real)
     char *temp;
     char *real;
 {
-    char bak[50];
+    char *bak;
     struct stat statbuf;
-    char rcs[PATH_MAX];
-    
+    char *rcs;
+
     /* Set "x" bits if set in original. */
+    rcs = xmalloc (strlen (real) + sizeof (RCSEXT) + 10);
     (void) sprintf (rcs, "%s%s", real, RCSEXT);
     statbuf.st_mode = 0; /* in case rcs file doesn't exist, but it should... */
-    (void) CVS_STAT (rcs, &statbuf);
+    if (CVS_STAT (rcs, &statbuf) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot stat %s", rcs);
+    free (rcs);
 
     if (chmod (temp, 0444 | (statbuf.st_mode & 0111)) < 0)
 	error (0, errno, "warning: cannot chmod %s", temp);
+    bak = xmalloc (strlen (real) + sizeof (BAKPREFIX) + 10);
     (void) sprintf (bak, "%s%s", BAKPREFIX, real);
-    (void) unlink_file (bak);		/* rm .#loginfo */
-    (void) CVS_RENAME (real, bak);		/* mv loginfo .#loginfo */
-    (void) CVS_RENAME (temp, real);		/* mv "temp" loginfo */
+
+    /* rm .#loginfo */
+    if (unlink_file (bak) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot remove %s", bak);
+
+    /* mv loginfo .#loginfo */
+    if (CVS_RENAME (real, bak) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot rename %s to %s", real, bak);
+
+    /* mv "temp" loginfo */
+    if (CVS_RENAME (temp, real) < 0
+	&& !existence_error (errno))
+	error (0, errno, "cannot rename %s to %s", temp, real);
+
+    free (bak);
 }
 
 const char *const init_usage[] = {
     "Usage: %s %s\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -621,6 +825,8 @@ init (argc, argv)
     char *info;
     /* Name of ,v file for this administrative file.  */
     char *info_v;
+    /* Exit status.  */
+    int err;
 
     const struct admin_file *fileptr;
 
@@ -652,7 +858,10 @@ init (argc, argv)
     strcat (adm, CVSROOTADM);
     mkdir_if_needed (adm);
 
-    /* This is needed by the call to "ci" below.  */
+    /* This is needed because we pass "fileptr->filename" not "info"
+       to add_rcs_file below.  I think this would be easy to change,
+       thus nuking the need for CVS_CHDIR here, but I haven't looked
+       closely (e.g. see wrappers calls within add_rcs_file).  */
     if ( CVS_CHDIR (adm) < 0)
 	error (1, errno, "cannot change to directory %s", adm);
 
@@ -689,16 +898,19 @@ init (argc, argv)
 		if (fclose (fp) < 0)
 		    error (1, errno, "cannot close %s", info);
 	    }
-	    /* Now check the file in.  FIXME: we could be using
-	       add_rcs_file from import.c which is faster (if it were
-	       tweaked slightly).  */
-	    run_setup ("%s%s -x,v/ -q -u -t-", Rcsbin, RCS_CI);
-	    run_args ("-minitial checkin of %s", fileptr->filename);
-	    run_arg (fileptr->filename);
-	    retcode = run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL);
+	    /* The message used to say " of " and fileptr->filename after
+	       "initial checkin" but I fail to see the point as we know what
+	       file it is from the name.  */
+	    retcode = add_rcs_file ("initial checkin", info_v,
+				    fileptr->filename, "1.1", NULL,
+
+				    /* No vendor branch.  */
+				    NULL, NULL, 0, NULL,
+
+				    NULL, 0, NULL);
 	    if (retcode != 0)
-		error (1, retcode == -1 ? errno : 0,
-		       "failed to check in %s", info);
+		/* add_rcs_file already printed an error message.  */
+		err = 1;
 	}
     }
 
