@@ -3,7 +3,7 @@
  *    Copyright (c) 1989-1992, Brian Berliner
  *
  *    You may distribute under the terms of the GNU General Public License
- *    as specified in the README file that comes with the CVS 1.4 kit.
+ *    as specified in the README file that comes with the CVS source distribution.
  *
  * Modules
  *
@@ -20,6 +20,7 @@
  *	command line.
  */
 
+#include <assert.h>
 #include "cvs.h"
 #include "savecwd.h"
 
@@ -36,8 +37,15 @@
 
 struct sortrec
 {
+    /* Name of the module, malloc'd.  */
     char *modname;
+    /* If Status variable is set, this is either def_status or the malloc'd
+       name of the status.  If Status is not set, the field is left
+       uninitialized.  */
     char *status;
+    /* Pointer to a malloc'd array which contains (1) the raw contents
+       of the options and arguments, excluding comments, (2) a '\0',
+       and (3) the storage for the "comment" field.  */
     char *rest;
     char *comment;
 };
@@ -54,18 +62,21 @@ static void save_d PROTO((char *k, int ks, char *d, int ds));
 DBM *
 open_module ()
 {
-    char mfile[PATH_MAX];
+    char *mfile;
+    DBM *retval;
 
     if (CVSroot_original == NULL)
     {
-	(void) fprintf (stderr, 
-			"%s: must set the CVSROOT environment variable\n",
-			program_name);
-	error (1, 0, "or specify the '-d' option to %s", program_name);
+	error (0, 0, "must set the CVSROOT environment variable");
+	error (1, 0, "or specify the '-d' global option");
     }
+    mfile = xmalloc (strlen (CVSroot_directory) + sizeof (CVSROOTADM)
+		     + sizeof (CVSROOTADM_MODULES) + 20);
     (void) sprintf (mfile, "%s/%s/%s", CVSroot_directory,
 		    CVSROOTADM, CVSROOTADM_MODULES);
-    return (dbm_open (mfile, O_RDONLY, 0666));
+    retval = dbm_open (mfile, O_RDONLY, 0666);
+    free (mfile);
+    return retval;
 }
 
 /*
@@ -104,23 +115,28 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     char *tag_prog = NULL;
     char *update_prog = NULL;
     struct saved_cwd cwd;
+    int cwd_saved = 0;
     char *line;
     int modargc;
     int xmodargc;
     char **modargv;
-    char *xmodargv[MAXFILEPERDIR];
-    char *value;
-    char *zvalue;
+    char **xmodargv;
+    /* Found entry from modules file, including options and such.  */
+    char *value = NULL;
+    char *zvalue = NULL;
     char *mwhere = NULL;
     char *mfile = NULL;
     char *spec_opt = NULL;
-    char xvalue[PATH_MAX];
+    char *xvalue = NULL;
     int alias = 0;
     datum key, val;
     char *cp;
     int c, err = 0;
+    int nonalias_opt = 0;
 
 #ifdef SERVER_SUPPORT
+    int restore_server_dir = 0;
+    char *server_dir_to_restore = NULL;
     if (trace)
     {
 	char *buf;
@@ -133,8 +149,8 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 		       + strlen (msg)
 		       + (where ? strlen (where) : 0)
 		       + (extra_arg ? strlen (extra_arg) : 0));
-	sprintf (buf, "%c-> do_module (%s, %s, %s, %s)\n",
-		 (server_active) ? 'S' : ' ',
+	sprintf (buf, "%s-> do_module (%s, %s, %s, %s)\n",
+		 CLIENT_SERVER_STR,
 		 mname, msg, where ? where : "",
 		 extra_arg ? extra_arg : "");
 	cvs_outerr (buf, 0);
@@ -146,11 +162,11 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     if (mname[0] == '!' && mname[1] != '\0')
     {
 	ign_dir_add (mname+1);
-	return(err);
+	goto do_module_return;
     }
 
     /* strip extra stuff from the module name */
-    strip_path (mname);
+    strip_trailing_slashes (mname);
 
     /*
      * Look up the module using the following scheme:
@@ -178,13 +194,13 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	{
 	    do
 		*cp-- = '\0';
-	    while (isspace (*cp));
+	    while (isspace ((unsigned char) *cp));
 	}
 	else
 	{
 	    /* Always strip trailing spaces */
 	    cp = strchr (val.dptr, '\0');
-	    while (cp > val.dptr && isspace(*--cp))
+	    while (cp > val.dptr && isspace ((unsigned char) *--cp))
 		*cp = '\0';
 	}
 
@@ -194,13 +210,16 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     }
     else
     {
-	char file[PATH_MAX];
-	char attic_file[PATH_MAX];
+	char *file;
+	char *attic_file;
 	char *acp;
+	int is_found = 0;
 
 	/* check to see if mname is a directory or file */
-
+	file = xmalloc (strlen (CVSroot_directory) + strlen (mname) + 10);
 	(void) sprintf (file, "%s/%s", CVSroot_directory, mname);
+	attic_file = xmalloc (strlen (CVSroot_directory) + strlen (mname)
+			      + sizeof (CVSATTIC) + sizeof (RCSEXT) + 15);
 	if ((acp = strrchr (mname, '/')) != NULL)
 	{
 	    *acp = '\0';
@@ -214,8 +233,10 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 
 	if (isdir (file))
 	{
-	    value = mname;
-	    goto found;
+	    modargv = xmalloc (sizeof (*modargv));
+	    modargv[0] = xstrdup (mname);
+	    modargc = 1;
+	    is_found = 1;
 	}
 	else
 	{
@@ -225,12 +246,12 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 		/* if mname was a file, we have to split it into "dir file" */
 		if ((cp = strrchr (mname, '/')) != NULL && cp != mname)
 		{
-		    char *slashp;
-
-		    /* put the ' ' in a copy so we don't mess up the original */
-		    value = strcpy (xvalue, mname);
-		    slashp = strrchr (value, '/');
-		    *slashp = ' ';
+		    modargv = xmalloc (2 * sizeof (*modargv));
+		    modargv[0] = xmalloc (strlen (mname) + 2);
+		    strncpy (modargv[0], mname, cp - mname);
+		    modargv[0][cp - mname] = '\0';
+		    modargv[1] = xstrdup (cp + 1);
+		    modargc = 2;
 		}
 		else
 		{
@@ -242,19 +263,54 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 		    if (cp == mname)
 		    {
 			/* drop the leading / if specified */
-			value = strcpy (xvalue, ". ");
-			(void) strcat (xvalue, mname + 1);
+			modargv = xmalloc (2 * sizeof (*modargv));
+			modargv[0] = xstrdup (".");
+			modargv[1] = xstrdup (mname + 1);
+			modargc = 2;
 		    }
 		    else
 		    {
 			/* otherwise just copy it */
-			value = strcpy (xvalue, ". ");
-			(void) strcat (xvalue, mname);
+			modargv = xmalloc (2 * sizeof (*modargv));
+			modargv[0] = xstrdup (".");
+			modargv[1] = xstrdup (mname);
+			modargc = 2;
 		    }
 		}
-		goto found;
+		is_found = 1;
 	    }
 	}
+	free (attic_file);
+	free (file);
+
+	if (is_found)
+	{
+	    assert (value == NULL);
+
+	    /* OK, we have now set up modargv with the actual
+	       file/directory we want to work on.  We duplicate a
+	       small amount of code here because the vast majority of
+	       the code after the "found" label does not pertain to
+	       the case where we found a file/directory rather than
+	       finding an entry in the modules file.  */
+	    if (save_cwd (&cwd))
+		error_exit ();
+	    cwd_saved = 1;
+
+	    err += callback_proc (&modargc, modargv, where, mwhere, mfile,
+				  shorten,
+				  local_specified, mname, msg);
+
+	    free_names (&modargc, modargv);
+
+	    /* cd back to where we started.  */
+	    if (restore_cwd (&cwd, NULL))
+		error_exit ();
+	    free_cwd (&cwd);
+	    cwd_saved = 0;
+
+	    goto do_module_return;
+	}
     }
 
     /* look up everything to the first / as a module */
@@ -284,7 +340,7 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	    {
 		do
 		    *cp2-- = '\0';
-		while (isspace (*cp2));
+		while (isspace ((unsigned char) *cp2));
 	    }
 	    value = val.dptr;
 
@@ -305,9 +361,7 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     /* if we got here, we couldn't find it using our search, so give up */
     error (0, 0, "cannot find module `%s' - ignored", mname);
     err++;
-    if (mwhere)
-	free (mwhere);
-    return (err);
+    goto do_module_return;
 
 
     /*
@@ -318,10 +372,12 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 
     /* remember where we start */
     if (save_cwd (&cwd))
-	exit (EXIT_FAILURE);
+	error_exit ();
+    cwd_saved = 1;
 
     /* copy value to our own string since if we go recursive we'll be
        really screwed if we do another dbm lookup */
+    assert (value != NULL);
     zvalue = xstrdup (value);
     value = zvalue;
 
@@ -332,7 +388,7 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	spec_opt = cp + 1;		/* save the options for later */
 
 	if (cp != value)		/* strip whitespace if necessary */
-	    while (isspace (*--cp))
+	    while (isspace ((unsigned char) *--cp))
 		*cp = '\0';
 
 	if (cp == value)
@@ -364,22 +420,12 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	    }
 	    if (!isfile (CVSADM))
 	    {
-		char nullrepos[PATH_MAX];
+		char *nullrepos;
 
-		(void) sprintf (nullrepos, "%s/%s/%s", CVSroot_directory,
-				CVSROOTADM, CVSNULLREPOS);
-		if (!isfile (nullrepos))
-		{
-		    mode_t omask;
-		    omask = umask (cvsumask);
-		    (void) CVS_MKDIR (nullrepos, 0777);
-		    (void) umask (omask);
-		}
-		if (!isdir (nullrepos))
-		    error (1, 0, "there is no repository %s", nullrepos);
+		nullrepos = emptydir_name ();
 
 		Create_Admin (".", dir,
-			      nullrepos, (char *) NULL, (char *) NULL);
+			      nullrepos, (char *) NULL, (char *) NULL, 0, 0);
 		if (!noexec)
 		{
 		    FILE *fp;
@@ -392,6 +438,7 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 			server_set_entstat (dir, nullrepos);
 #endif
 		}
+		free (nullrepos);
 	    }
 	  out:
 	    goto do_special;
@@ -416,13 +463,13 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     (void) sprintf (line, "%s %s", "XXX", value);
 
     /* turn the line into an argv[] array */
-    line2argv (&xmodargc, xmodargv, line);
+    line2argv (&xmodargc, &xmodargv, line, " \t");
     free (line);
     modargc = xmodargc;
     modargv = xmodargv;
 
     /* parse the args */
-    optind = 1;
+    optind = 0;
     while ((c = getopt (modargc, modargv, CVSMODULE_OPTS)) != -1)
     {
 	switch (c)
@@ -431,38 +478,51 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 		alias = 1;
 		break;
 	    case 'd':
+		nonalias_opt = 1;
 		if (mwhere)
 		    free (mwhere);
 		mwhere = xstrdup (optarg);
 		break;
 	    case 'i':
-		checkin_prog = optarg;
+		nonalias_opt = 1;
+		if (checkin_prog)
+		    free (checkin_prog);
+		checkin_prog = xstrdup (optarg);
 		break;
 	    case 'l':
+		nonalias_opt = 1;
 		local_specified = 1;
 		break;
 	    case 'o':
-		checkout_prog = optarg;
+		nonalias_opt = 1;
+		if (checkout_prog)
+		    free (checkout_prog);
+		checkout_prog = xstrdup (optarg);
 		break;
 	    case 'e':
-		export_prog = optarg;
+		nonalias_opt = 1;
+		if (export_prog)
+		    free (export_prog);
+		export_prog = xstrdup (optarg);
 		break;
 	    case 't':
-		tag_prog = optarg;
+		nonalias_opt = 1;
+		if (tag_prog)
+		    free (tag_prog);
+		tag_prog = xstrdup (optarg);
 		break;
 	    case 'u':
-		update_prog = optarg;
+		nonalias_opt = 1;
+		if (update_prog)
+		    free (update_prog);
+		update_prog = xstrdup (optarg);
 		break;
 	    case '?':
 		error (0, 0,
 		       "modules file has invalid option for key %s value %s",
 		       key.dptr, val.dptr);
 		err++;
-		if (mwhere)
-		    free (mwhere);
-		free (zvalue);
-		free_cwd (&cwd);
-		return (err);
+		goto do_module_return;
 	}
     }
     modargc -= optind;
@@ -470,11 +530,20 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     if (modargc == 0)
     {
 	error (0, 0, "modules file missing directory for module %s", mname);
-	if (mwhere)
-	    free (mwhere);
-	free (zvalue);
-	free_cwd (&cwd);
-	return (++err);
+	++err;
+	goto do_module_return;
+    }
+
+    if (alias && nonalias_opt)
+    {
+	/* The documentation has never said it is legal to specify
+	   -a along with another option.  And I believe that in the past
+	   CVS has ignored the options other than -a, more or less, in this
+	   situation.  */
+	error (0, 0, "\
+-a cannot be specified in the modules file along with other options");
+	++err;
+	goto do_module_return;
     }
 
     /* if this was an alias, call ourselves recursively for each module */
@@ -493,23 +562,23 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 				  where, shorten, local_specified,
 				  run_module_prog, extra_arg);
 	}
-	if (mwhere)
-	    free (mwhere);
-	free (zvalue);
-	free_cwd (&cwd);
-	return (err);
+	goto do_module_return;
+    }
+
+    if (mfile != NULL && modargc > 1)
+    {
+	error (0, 0, "\
+module `%s' is a request for a file in a module which is not a directory",
+	       mname);
+	++err;
+	goto do_module_return;
     }
 
     /* otherwise, process this module */
     err += callback_proc (&modargc, modargv, where, mwhere, mfile, shorten,
 			  local_specified, mname, msg);
 
-#if 0
-    /* FIXME: I've fixed this so that the correct arguments are called, 
-       but now this fails because there is code below this point that
-       uses optarg values extracted from the arg vector. */
     free_names (&xmodargc, xmodargv);
-#endif
 
     /* if there were special include args, process them now */
 
@@ -519,6 +588,40 @@ do_module (db, mname, m_type, msg, callback_proc, where,
     if (local_specified)
 	spec_opt = NULL;
 
+#ifdef SERVER_SUPPORT
+    /* We want to check out into the directory named by the module.
+       So we set a global variable which tells the server to glom that
+       directory name onto the front.  A cleaner approach would be some
+       way of passing it down to the recursive call, through the
+       callback_proc, to start_recursion, and then into the update_dir in
+       the struct file_info.  That way the "Updating foo" message could
+       print the actual directory we are checking out into.
+
+       For local CVS, this is handled by the chdir call above
+       (directly or via the callback_proc).  */
+    if (server_active && spec_opt != NULL)
+    {
+	char *change_to;
+
+	change_to = where ? where : (mwhere ? mwhere : mname);
+	server_dir_to_restore = server_dir;
+	restore_server_dir = 1;
+	server_dir =
+	    xmalloc ((server_dir_to_restore != NULL
+		      ? strlen (server_dir_to_restore)
+		      : 0)
+		     + strlen (change_to)
+		     + 5);
+	server_dir[0] = '\0';
+	if (server_dir_to_restore != NULL)
+	{
+	    strcat (server_dir, server_dir_to_restore);
+	    strcat (server_dir, "/");
+	}
+	strcat (server_dir, change_to);
+    }
+#endif
+
     while (spec_opt != NULL)
     {
 	char *next_opt;
@@ -532,13 +635,13 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	    /* strip whitespace off the end */
 	    do
 		*cp = '\0';
-	    while (isspace (*--cp));
+	    while (isspace ((unsigned char) *--cp));
 	}
 	else
 	    next_opt = NULL;
 
 	/* strip whitespace from front */
-	while (isspace (*spec_opt))
+	while (isspace ((unsigned char) *spec_opt))
 	    spec_opt++;
 
 	if (*spec_opt == '\0')
@@ -551,6 +654,14 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	spec_opt = next_opt;
     }
 
+#ifdef SERVER_SUPPORT
+    if (server_active && restore_server_dir)
+    {
+	free (server_dir);
+	server_dir = server_dir_to_restore;
+    }
+#endif
+
     /* write out the checkin/update prog files if necessary */
 #ifdef SERVER_SUPPORT
     if (err == 0 && !noexec && m_type == CHECKOUT && server_expanding)
@@ -584,8 +695,9 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 
     /* cd back to where we started */
     if (restore_cwd (&cwd, NULL))
-	exit (EXIT_FAILURE);
+	error_exit ();
     free_cwd (&cwd);
+    cwd_saved = 0;
 
     /* run checkout or tag prog if appropriate */
     if (err == 0 && run_module_prog)
@@ -600,7 +712,7 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	     * if we can't find a matching program, just punt and use
 	     * whatever is specified in the modules file.
 	     */
-	    char real_prog[PATH_MAX];
+	    char *real_prog = NULL;
 	    char *prog = (m_type == TAG ? tag_prog :
 			  (m_type == CHECKOUT ? checkout_prog : export_prog));
 	    char *real_where = (where != NULL ? where : mwhere);
@@ -608,6 +720,8 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 
 	    if ((*prog != '/') && (*prog != '.'))
 	    {
+		real_prog = xmalloc (strlen (real_where) + strlen (prog)
+				     + 10);
 		(void) sprintf (real_prog, "%s/%s", real_where, prog);
 		if (isfile (real_prog))
 		    prog = real_prog;
@@ -617,29 +731,49 @@ do_module (db, mname, m_type, msg, callback_proc, where,
 	    expanded_path = expand_path (prog, "modules", 0);
 	    if (expanded_path != NULL)
 	    {
-		run_setup ("%s %s", expanded_path, real_where);
+		run_setup (expanded_path);
+		run_arg (real_where);
 
 		if (extra_arg)
 		    run_arg (extra_arg);
 
 		if (!quiet)
 		{
-		    (void) printf ("%s %s: Executing '", program_name,
-				   command_name);
+		    cvs_output (program_name, 0);
+		    cvs_output (" ", 1);
+		    cvs_output (command_name, 0);
+		    cvs_output (": Executing '", 0);
 		    run_print (stdout);
-		    (void) printf ("'\n");
+		    cvs_output ("'\n", 0);
 		}
 		err += run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL);
 		free (expanded_path);
 	    }
+	    free (real_prog);
 	}
     }
 
+ do_module_return:
     /* clean up */
     if (mwhere)
 	free (mwhere);
-    free (zvalue);
+    if (checkin_prog)
+	free (checkin_prog);
+    if (checkout_prog)
+	free (checkout_prog);
+    if (export_prog)
+	free (export_prog);
+    if (tag_prog)
+	free (tag_prog);
+    if (update_prog)
+	free (update_prog);
+    if (cwd_saved)
+	free_cwd (&cwd);
+    if (zvalue != NULL)
+	free (zvalue);
 
+    if (xvalue != NULL)
+	free (xvalue);
     return (err);
 }
 
@@ -728,15 +862,15 @@ save_d (k, ks, d, ds)
     cp = d;
     *(cp + ds) = '\0';	/* Assumes an extra byte at end of static dbm buffer */
 
-    while (isspace (*cp))
+    while (isspace ((unsigned char) *cp))
 	cp++;
     /* Turn <spaces> into one ' ' -- makes the rest of this routine simpler */
     while (*cp)
     {
-	if (isspace (*cp))
+	if (isspace ((unsigned char) *cp))
 	{
 	    *cp2++ = ' ';
-	    while (isspace (*cp))
+	    while (isspace ((unsigned char) *cp))
 		cp++;
 	}
 	else
@@ -749,15 +883,19 @@ save_d (k, ks, d, ds)
     {
 	s_rec->status = def_status;
 
-	/* Minor kluge, but general enough to maintain */
 	for (cp = s_rec->rest; (cp2 = strchr (cp, '-')) != NULL; cp = ++cp2)
 	{
 	    if (*(cp2 + 1) == 's' && *(cp2 + 2) == ' ')
 	    {
-		s_rec->status = (cp2 += 3);
-		while (*cp2 != ' ')
+		char *status_start;
+
+		cp2 += 3;
+		status_start = cp2;
+		while (*cp2 != ' ' && *cp2 != '\0')
 		    cp2++;
-		*cp2++ = '\0';
+		s_rec->status = xmalloc (cp2 - status_start + 1);
+		strncpy (s_rec->status, status_start, cp2 - status_start);
+		s_rec->status[cp2 - status_start] = '\0';
 		cp = cp2;
 		break;
 	    }
@@ -794,8 +932,7 @@ cat_module (status)
     int moduleargc;
     struct sortrec *s_h;
     char *cp, *cp2, **argv;
-    char *line;
-    char *moduleargv[MAXFILEPERDIR];
+    char **moduleargv;
 
     Status = status;
 
@@ -810,6 +947,8 @@ cat_module (status)
 	    save_d (key.dptr, key.dsize, val.dptr, val.dsize);
     }
 
+    close_module (db);
+
     /* Sort the list as requested */
     qsort ((PTR) s_head, s_count, sizeof (struct sortrec), sort_order);
 
@@ -821,19 +960,25 @@ cat_module (status)
     fill = cols - (indent + 2);
     for (s_h = s_head, i = 0; i < s_count; i++, s_h++)
     {
+	char *line;
+
 	/* Print module name (and status, if wanted) */
-	(void) printf ("%-12s", s_h->modname);
+	line = xmalloc (strlen (s_h->modname) + 15);
+	sprintf (line, "%-12s", s_h->modname);
+	cvs_output (line, 0);
+	free (line);
 	if (status)
 	{
-	    (void) printf (" %-11s", s_h->status);
-	    if (s_h->status != def_status)
-		*(s_h->status + strlen (s_h->status)) = ' ';
+	    line = xmalloc (strlen (s_h->status) + 15);
+	    sprintf (line, " %-11s", s_h->status);
+	    cvs_output (line, 0);
+	    free (line);
 	}
 
+	line = xmalloc (strlen (s_h->modname) + strlen (s_h->rest) + 15);
 	/* Parse module file entry as command line and print options */
-	line = xmalloc (strlen (s_h->modname) + strlen (s_h->rest) + 10);
 	(void) sprintf (line, "%s %s", s_h->modname, s_h->rest);
-	line2argv (&moduleargc, moduleargv, line);
+	line2argv (&moduleargc, &moduleargv, line, " \t");
 	free (line);
 	argc = moduleargc;
 	argv = moduleargv;
@@ -846,17 +991,28 @@ cat_module (status)
 	    {
 		if (c == 'a' || c == 'l')
 		{
-		    (void) printf (" -%c", c);
+		    char buf[5];
+
+		    sprintf (buf, " -%c", c);
+		    cvs_output (buf, 0);
 		    wid += 3;		/* Could just set it to 3 */
 		}
 		else
 		{
+		    char buf[10];
+
 		    if (strlen (optarg) + 4 + wid > (unsigned) fill)
 		    {
-			(void) printf ("\n%*s", indent, "");
+			int j;
+
+			cvs_output ("\n", 1);
+			for (j = 0; j < indent; ++j)
+			    cvs_output (" ", 1);
 			wid = 0;
 		    }
-		    (void) printf (" -%c %s", c, optarg);
+		    sprintf (buf, " -%c ", c);
+		    cvs_output (buf, 0);
+		    cvs_output (optarg, 0);
 		    wid += strlen (optarg) + 4;
 		}
 	    }
@@ -869,21 +1025,31 @@ cat_module (status)
 	{
 	    if (strlen (*argv) + wid > (unsigned) fill)
 	    {
-		(void) printf ("\n%*s", indent, "");
+		int j;
+
+		cvs_output ("\n", 1);
+		for (j = 0; j < indent; ++j)
+		    cvs_output (" ", 1);
 		wid = 0;
 	    }
-	    (void) printf (" %s", *argv);
+	    cvs_output (" ", 1);
+	    cvs_output (*argv, 0);
 	    wid += strlen (*argv) + 1;
 	}
-	(void) printf ("\n");
+	cvs_output ("\n", 1);
 
 	/* Format the comment field -- save_d (), compressed spaces */
 	for (cp2 = cp = s_h->comment; *cp; cp2 = cp)
 	{
-	    (void) printf ("%*s # ", indent, "");
+	    int j;
+
+	    for (j = 0; j < indent; ++j)
+		cvs_output (" ", 1);
+	    cvs_output (" # ", 0);
 	    if (strlen (cp2) < (unsigned) (fill - 2))
 	    {
-		(void) printf ("%s\n", cp2);
+		cvs_output (cp2, 0);
+		cvs_output ("\n", 1);
 		break;
 	    }
 	    cp += fill - 2;
@@ -891,14 +1057,22 @@ cat_module (status)
 		cp--;
 	    if (cp == cp2)
 	    {
-		(void) printf ("%s\n", cp2);
+		cvs_output (cp2, 0);
+		cvs_output ("\n", 1);
 		break;
 	    }
 
 	    *cp++ = '\0';
-	    (void) printf ("%s\n", cp2);
+	    cvs_output (cp2, 0);
+	    cvs_output ("\n", 1);
 	}
 
 	free_names(&moduleargc, moduleargv);
+	/* FIXME-leak: here is where we would free s_h->modname, s_h->rest,
+	   and if applicable, s_h->status.  Not exactly a memory leak,
+	   in the sense that we are about to exit(), but may be worth
+	   noting if we ever do a multithreaded server or something of
+	   the sort.  */
     }
+    /* FIXME-leak: as above, here is where we would free s_head.  */
 }
