@@ -82,13 +82,8 @@ No_Difference (finfo, vers)
 	/* Need to call unlink myself because the noexec variable
 	 * has been set to 1.  */
 	if (trace)
-	    (void) fprintf (stderr, "%c-> unlink (%s)\n",
-#ifdef SERVER_SUPPORT
-			    (server_active) ? 'S' : ' ',
-#else
-			    ' ',
-#endif
-			    tocvsPath);
+	    (void) fprintf (stderr, "%s-> unlink (%s)\n",
+			    CLIENT_SERVER_STR, tocvsPath);
 	if ( CVS_UNLINK (tocvsPath) < 0)
 	    error (0, errno, "could not remove %s", tocvsPath);
     }
