@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Tag
  * 
@@ -59,20 +59,22 @@ static List *tlist;
 
 static const char *const tag_usage[] =
 {
-    "Usage: %s %s [-lRF] [-b] [-d] [-c] [-r tag|-D date] tag [files...]\n",
+    "Usage: %s %s [-lRF] [-b] [-d] [-c] [-r rev|-D date] tag [files...]\n",
     "\t-l\tLocal directory only, not recursive.\n",
     "\t-R\tProcess directories recursively.\n",
     "\t-d\tDelete the given tag.\n",
-    "\t-[rD]\tExisting tag or date.\n",
+    "\t-r rev\tExisting revision/tag.\n",
+    "\t-D\tExisting date.\n",
     "\t-f\tForce a head revision if specified tag not found.\n",
     "\t-b\tMake the tag a \"branch\" tag, allowing concurrent development.\n",
     "\t-F\tMove tag if it already exists.\n",
     "\t-c\tCheck that working files are unmodified.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
 int
-tag (argc, argv)
+cvstag (argc, argv)
     int argc;
     char **argv;
 {
@@ -82,8 +84,8 @@ tag (argc, argv)
     if (argc == -1)
 	usage (tag_usage);
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "FQqlRcdr:D:bf")) != -1)
+    optind = 0;
+    while ((c = getopt (argc, argv, "+FQqlRcdr:D:bf")) != -1)
     {
 	switch (c)
 	{
@@ -156,6 +158,8 @@ tag (argc, argv)
 	
 	ign_setup ();
 
+	if (!force_tag_match)
+	    send_arg ("-f");
 	if (local)
 	    send_arg("-l");
 	if (delete_flag)
@@ -174,11 +178,13 @@ tag (argc, argv)
 
 	send_arg (symtag);
 
+	send_files (argc, argv, local, 0,
+
+		    /* I think the -c case is like "cvs status", in
+		       which we really better be correct rather than
+		       being fast; it is just too confusing otherwise.  */
+		    check_uptodate ? 0 : SEND_NO_CONTENTS);
 	send_file_names (argc, argv, SEND_EXPAND_WILD);
-	/* FIXME:  We shouldn't have to send current files, but I'm not sure
-	   whether it works.  So send the files --
-	   it's slower but it works.  */
-	send_files (argc, argv, local, 0);
 	send_to_server ("tag\012", 0);
         return get_responses_and_close ();
     }
@@ -267,8 +273,17 @@ check_fileproc (callerdat, finfo)
 	    error (0, 0, "nothing known about %s", finfo->file);
 	return (1);
     }
-    p->data = RCS_getversion(vers->srcfile, numtag, date, force_tag_match,
-			     (int *) NULL);
+
+    /* Here we duplicate the calculation in tag_fileproc about which
+       version we are going to tag.  There probably are some subtle races
+       (e.g. numtag is "foo" which gets moved between here and
+       tag_fileproc).  */
+    if (numtag == NULL && date == NULL)
+	p->data = xstrdup (vers->vn_user);
+    else
+	p->data = RCS_getversion (vers->srcfile, numtag, date,
+				  force_tag_match, NULL);
+
     if (p->data != NULL)
     {
         int addit = 1;
@@ -280,9 +295,16 @@ check_fileproc (callerdat, finfo)
         {
             if (delete_flag)
             {
+		/* Deleting a tag which did not exist is a noop and
+		   should not be logged.  */
                 addit = 0;
             }
         }
+	else if (delete_flag)
+	{
+	    free (p->data);
+	    p->data = xstrdup (oversion);
+	}
         else if (strcmp(oversion, p->data) == 0)
         {
             addit = 0;
@@ -350,7 +372,7 @@ pretag_proc(repository, filter)
         s = xstrdup(filter);
         for (cp=s; *cp; cp++)
         {
-            if (isspace(*cp))
+            if (isspace ((unsigned char) *cp))
             {
                 *cp = '\0';
                 break;
@@ -364,13 +386,12 @@ pretag_proc(repository, filter)
         }
         free(s);
     }
-    run_setup("%s %s %s %s",
-              filter,
-              symtag,
-              delete_flag ? "del" : force_tag_move ? "mov" : "add",
-              repository);
+    run_setup (filter);
+    run_arg (symtag);
+    run_arg (delete_flag ? "del" : force_tag_move ? "mov" : "add");
+    run_arg (repository);
     walklist(tlist, pretag_list_proc, NULL);
-    return (run_exec(RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL|RUN_REALLY));
+    return (run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL));
 }
 
 static void
@@ -430,7 +451,13 @@ tag_fileproc (callerdat, finfo)
     /* Lock the directory if it is not already locked.  We can't rely
        on tag_dirproc because it won't handle the case where the user
        specifies a list of files on the command line.  */
-    tag_lockdir (finfo->repository);
+    /* We do not need to acquire a full write lock for the tag operation:
+       the revisions are obtained from the working directory, so we do not
+       require consistency across the entire repository.  However, we do
+       need to prevent simultaneous tag operations from interfering with
+       each other.  Therefore, we write lock each directory as we enter
+       it, and unlock it as we leave it.  */
+    lock_dir_for_write (finfo->repository);
 
     vers = Version_TS (finfo, NULL, NULL, NULL, 0, 0);
 
@@ -468,7 +495,7 @@ tag_fileproc (callerdat, finfo)
 	}
 	free (version);
 
-	if ((retcode = RCS_deltag(vers->srcfile, symtag, 1)) != 0) 
+	if ((retcode = RCS_deltag(vers->srcfile, symtag)) != 0) 
 	{
 	    if (!quiet)
 		error (0, retcode == -1 ? errno : 0,
@@ -477,11 +504,14 @@ tag_fileproc (callerdat, finfo)
 	    freevers_ts (&vers);
 	    return (1);
 	}
+	RCS_rewrite (vers->srcfile, NULL, NULL);
 
 	/* warm fuzzies */
 	if (!really_quiet)
 	{
-	    (void) printf ("D %s\n", finfo->fullname);
+	    cvs_output ("D ", 2);
+	    cvs_output (finfo->fullname, 0);
+	    cvs_output ("\n", 1);
 	}
 
 	freevers_ts (&vers);
@@ -540,7 +570,7 @@ tag_fileproc (callerdat, finfo)
 			       (int *) NULL);
     if (oversion != NULL)
     {
-	int isbranch = RCS_isbranch (finfo->rcs, symtag);
+	int isbranch = RCS_nodeisbranch (finfo->rcs, symtag);
 
 	/*
 	 * if versions the same and neither old or new are branches don't have 
@@ -556,12 +586,19 @@ tag_fileproc (callerdat, finfo)
 	if (!force_tag_move)
 	{
 	    /* we're NOT going to move the tag */
-	    (void) printf ("W %s", finfo->fullname);
-
-	    (void) printf (" : %s already exists on %s %s", 
-			   symtag, isbranch ? "branch" : "version", oversion);
-	    (void) printf (" : NOT MOVING tag to %s %s\n", 
-			   branch_mode ? "branch" : "version", rev);
+	    cvs_output ("W ", 2);
+	    cvs_output (finfo->fullname, 0);
+	    cvs_output (" : ", 0);
+	    cvs_output (symtag, 0);
+	    cvs_output (" already exists on ", 0);
+	    cvs_output (isbranch ? "branch" : "version", 0);
+	    cvs_output (" ", 0);
+	    cvs_output (oversion, 0);
+	    cvs_output (" : NOT MOVING tag to ", 0);
+	    cvs_output (branch_mode ? "branch" : "version", 0);
+	    cvs_output (" ", 0);
+	    cvs_output (rev, 0);
+	    cvs_output ("\n", 1);
 	    free (oversion);
 	    freevers_ts (&vers);
 	    return (0);
@@ -577,11 +614,14 @@ tag_fileproc (callerdat, finfo)
 	freevers_ts (&vers);
 	return (1);
     }
+    RCS_rewrite (vers->srcfile, NULL, NULL);
 
     /* more warm fuzzies */
     if (!really_quiet)
     {
-	(void) printf ("T %s\n", finfo->fullname);
+	cvs_output ("T ", 2);
+	cvs_output (finfo->fullname, 0);
+	cvs_output ("\n", 1);
     }
 
     if (nversion != NULL)
@@ -602,7 +642,7 @@ tag_filesdoneproc (callerdat, err, repos, update_dir, entries)
     char *update_dir;
     List *entries;
 {
-    tag_unlockdir ();
+    Lock_Cleanup ();
 
     return (err);
 }
@@ -624,66 +664,6 @@ tag_dirproc (callerdat, dir, repos, update_dir, entries)
     return (R_PROCESS);
 }
 
-/* We do not need to acquire a full write lock for the tag operation:
-   the revisions are obtained from the working directory, so we do not
-   require consistency across the entire repository.  However, we do
-   need to prevent simultaneous tag operations from interfering with
-   each other.  Therefore, we write lock each directory as we enter
-   it, and unlock it as we leave it.
-
-   In the rtag case, it would be nice to provide consistency with
-   respect to commits; however CVS lacks the infrastructure to do that
-   (see Concurrency in cvs.texinfo and comment in do_recursion).  We
-   can and will prevent simultaneous tag operations from interfering
-   with each other, by write locking each directory as we enter it,
-   and unlocking it as we leave it.  */
-static char *locked_dir;
-static List *locked_list;
-
-/*
- * Lock the directory for a tag operation.  This is also called by the
- * rtag code.
- */
-void
-tag_lockdir (repository)
-     char *repository;
-{
-    if (repository != NULL
-	&& (locked_dir == NULL
-	    || strcmp (locked_dir, repository) != 0))
-    {
-        Node *node;
-
-	if (locked_dir != NULL)
-	    tag_unlockdir ();
-
-	locked_dir = xstrdup (repository);
-	locked_list = getlist ();
-	node = getnode ();
-	node->type = LOCK;
-	node->key = xstrdup (repository);
-	(void) addnode (locked_list, node);
-	Writer_Lock (locked_list);
-    }
-}
-
-/*
- * Unlock the directory for a tag operation.  This is also called by
- * the rtag code.
- */
-void
-tag_unlockdir ()
-{
-    if (locked_dir != NULL)
-    {
-        Lock_Cleanup ();
-	dellist (&locked_list);
-	free (locked_dir);
-	locked_dir = NULL;
-	locked_list = NULL;
-    }
-}
-
 /* Code relating to the val-tags file.  Note that this file has no way
    of knowing when a tag has been deleted.  The problem is that there
    is no way of knowing whether a tag still exists somewhere, when we
@@ -772,12 +752,12 @@ tag_check_valid (name, argc, argv, local, aflag, repository)
     int which;
 
     /* Numeric tags require only a syntactic check.  */
-    if (isdigit (name[0]))
+    if (isdigit ((unsigned char) name[0]))
     {
 	char *p;
 	for (p = name; *p != '\0'; ++p)
 	{
-	    if (!(isdigit (*p) || *p == '.'))
+	    if (!(isdigit ((unsigned char) *p) || *p == '.'))
 		error (1, 0, "\
 Numeric tag %s contains characters other than digits and '.'", name);
 	}
@@ -789,6 +769,11 @@ Numeric tag %s contains characters other than digits and '.'", name);
 	|| strcmp (name, TAG_HEAD) == 0)
 	return;
 
+    /* FIXME: This routine doesn't seem to do any locking whatsoever
+       (and it is called from places which don't have locks in place).
+       If two processes try to write val-tags at the same time, it would
+       seem like we are in trouble.  */
+
     mytag.dptr = name;
     mytag.dsize = strlen (name);
 
@@ -846,7 +831,7 @@ Numeric tag %s contains characters other than digits and '.'", name);
 	else
 	{
 	    if (save_cwd (&cwd))
-		exit (EXIT_FAILURE);
+		error_exit ();
 	    if ( CVS_CHDIR (repository) < 0)
 		error (1, errno, "cannot change to %s directory", repository);
 	}
@@ -923,12 +908,19 @@ tag_check_valid_join (join_tag, argc, argv, local, aflag, repository)
     s = strchr (c, ':');
     if (s != NULL)
     {
-        if (isdigit (join_tag[0]))
+        if (isdigit ((unsigned char) join_tag[0]))
 	    error (1, 0,
 		   "Numeric join tag %s may not contain a date specifier",
 		   join_tag);
 
         *s = '\0';
+	/* hmmm...  I think it makes sense to allow -j:<date>, but
+	 * for now this fixes a bug where CVS just spins and spins (I
+	 * think in the RCS code) looking for a zero length tag.
+	 */
+	if (!*c)
+	    error (1, 0,
+		   "argument to join may not contain a date specifier without a tag");
     }
 
     tag_check_valid (c, argc, argv, local, aflag, repository);
